# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for utils module
"""

import os
from math import sqrt

import pytest

from .. import mp as utils_mp
from ..misc import null_context

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

pytestmark = pytest.mark.filterwarnings(
    'always:multiprocessing is currently not supported on Windws')


@pytest.mark.parametrize('verbose', [False, True, 'Test'])
@pytest.mark.parametrize('nproc', [1, 2])
def test_multiprocess_with_queues(capsys, nproc, verbose):
    inputs = [1, 4, 9, 16, 25]
    if os.name == 'nt' and nproc > 1:
        ctx = pytest.warns(UserWarning)
    else:
        ctx = null_context()

    with ctx:  # on windows, mp prints warning and falls back to nproc=1
        out = utils_mp.multiprocess_with_queues(
            nproc, sqrt, inputs, verbose=verbose,
        )
    assert out == [1, 2, 3, 4, 5]

    # assert progress bar prints correctly
    cap = capsys.readouterr()
    if verbose is True:
        assert cap.out.startswith('\rProcessing: ')
    elif verbose is False:
        assert cap.out == ''
    else:
        assert cap.out.startswith('\r{}: '.format(verbose))

    with pytest.raises(ValueError):
        utils_mp.multiprocess_with_queues(nproc, sqrt, [-1],
                                          verbose=verbose)


def test_multiprocess_with_queues_raise():
    with pytest.warns(DeprecationWarning):
        utils_mp.multiprocess_with_queues(1, sqrt, [1],
                                          raise_exceptions=True)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.nds2`
"""

import os

import pytest

from ...tests import mocks
from ...tests.mocks import mock
from ...tests.utils import skip_missing_dependency
from .. import nds2 as io_nds2

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def test_channel_type_find():
    """Test `gwpy.io.nds2.Nds2ChannelType` enum
    """
    # check that m-trend gets recognised properly
    a = io_nds2.Nds2ChannelType.find('m-trend')
    b = io_nds2.Nds2ChannelType.find('MTREND')
    assert a == b == io_nds2.Nds2ChannelType.MTREND
    # test unknown
    with pytest.raises(ValueError):
        io_nds2.Nds2ChannelType.find('blah')


def test_data_type_find():
    """Test `gwpy.io.nds2.Nds2DataType` enum
    """
    # test float
    assert io_nds2.Nds2DataType.find(float) == io_nds2.Nds2DataType.FLOAT64
    # test uint
    assert io_nds2.Nds2DataType.find('uint32') == \
        io_nds2.Nds2DataType.UINT32
    # test unknown
    with pytest.raises(TypeError):
        io_nds2.Nds2DataType.find('blah')


def test_parse_nds_env():
    """Test :func:`gwpy.io.nds2.parse_nds_env`
    """
    # test basic NDSSERVER usage
    os.environ['NDSSERVER'] = 'test1.ligo.org:80,test2.ligo.org:43'
    hosts = io_nds2.parse_nds_env()
    assert hosts == [('test1.ligo.org', 80), ('test2.ligo.org', 43)]

    # check that duplicates get removed
    os.environ['NDSSERVER'] = ('test1.ligo.org:80,test2.ligo.org:43,'
                               'test.ligo.org,test2.ligo.org:43')
    hosts = io_nds2.parse_nds_env()
    assert hosts == [('test1.ligo.org', 80), ('test2.ligo.org', 43),
                     ('test.ligo.org', None)]

    # check that a named environment variable works
    os.environ['TESTENV'] = 'test1.ligo.org:80,test2.ligo.org:43'
    hosts = io_nds2.parse_nds_env('TESTENV')
    assert hosts == [('test1.ligo.org', 80), ('test2.ligo.org', 43)]


def test_nds2_host_order():
    """Test :func:`gwpy.io.nds2.host_resolution_order`
    """
    # check None returns CIT
    hro = io_nds2.host_resolution_order(None, env=None)
    assert hro == [('nds.ligo.caltech.edu', 31200)]

    # check L1 returns (LLO, CIT)
    hro = io_nds2.host_resolution_order('L1', env=None)
    assert hro == [('nds.ligo-la.caltech.edu', 31200),
                   ('nds.ligo.caltech.edu', 31200)]

    # check NDSSERVER works
    os.environ['NDSSERVER'] = 'test1.ligo.org:80,test2.ligo.org:43'
    hro = io_nds2.host_resolution_order(None)
    assert hro == [('test1.ligo.org', 80), ('test2.ligo.org', 43),
                   ('nds.ligo.caltech.edu', 31200)]

    # check that NDSSERVER and an IFO spec works at the same time
    hro = io_nds2.host_resolution_order('L1')
    assert hro == [('test1.ligo.org', 80), ('test2.ligo.org', 43),
                   ('nds.ligo-la.caltech.edu', 31200),
                   ('nds.ligo.caltech.edu', 31200)]

    # test named environment variable
    os.environ['TESTENV'] = 'test1.ligo.org:80,test2.ligo.org:43'
    hro = io_nds2.host_resolution_order(None, env='TESTENV')
    assert hro == [('test1.ligo.org', 80), ('test2.ligo.org', 43),
                   ('nds.ligo.caltech.edu', 31200)]

    # test epoch='now' doesn't change anything
    os.environ.pop('NDSSERVER')
    hro = io_nds2.host_resolution_order('L1', epoch='now', env=None)
    assert hro == [('nds.ligo-la.caltech.edu', 31200),
                   ('nds.ligo.caltech.edu', 31200)]

    # test old epoch puts CIT ahead of LLO
    hro = io_nds2.host_resolution_order('L1', epoch='Jan 1 2015', env=None)
    assert hro == [('nds.ligo.caltech.edu', 31200),
                   ('nds.ligo-la.caltech.edu', 31200)]

    # test epoch doesn't operate with env
    hro = io_nds2.host_resolution_order('L1', epoch='now', env='TESTENV')
    assert hro == [('test1.ligo.org', 80), ('test2.ligo.org', 43),
                   ('nds.ligo-la.caltech.edu', 31200),
                   ('nds.ligo.caltech.edu', 31200)]

    # test warnings for unknown IFO
    with pytest.warns(UserWarning):
        hro = io_nds2.host_resolution_order('X1')
        assert hro == [('nds.ligo.caltech.edu', 31200)]


@skip_missing_dependency('nds2')
def test_connect():
    """Test :func:`gwpy.io.connect`
    """
    nds_connection = mocks.nds2_connection(host='nds.test.gwpy')
    with mock.patch('nds2.connection') as mock_connection:
        mock_connection.return_value = nds_connection
        conn = io_nds2.connect('nds.test.gwpy')
        assert conn.get_host() == 'nds.test.gwpy'
        assert conn.get_port() == 31200

    nds_connection = mocks.nds2_connection(host='nds2.test.gwpy',
                                           port=8088)
    with mock.patch('nds2.connection') as mock_connection:
        mock_connection.return_value = nds_connection
        conn = io_nds2.connect('nds2.test.gwpy')
        assert conn.get_host() == 'nds2.test.gwpy'
        assert conn.get_port() == 8088


def test_minute_trend_times():
    """Test :func:`gwpy.io.nds2.minute_trend_times`
    """
    assert io_nds2.minute_trend_times(0, 60) == (0, 60)
    assert io_nds2.minute_trend_times(1, 60) == (0, 60)
    assert io_nds2.minute_trend_times(0, 61) == (0, 120)
    assert io_nds2.minute_trend_times(59, 61) == (0, 120)
    assert (io_nds2.minute_trend_times(1167264018, 1198800018) ==
            (1167264000, 1198800060))



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.astro.range`
"""

import os.path

import pytest

from scipy import __version__ as scipy_version

from astropy import units

from ... import astro
from ...tests import utils
from ...timeseries import TimeSeries
from ...frequencyseries import FrequencySeries

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# -- test results -------------------------------------------------------------

# hack up constants that changed between astropy 1.3 and 2.0
# TODO: might want to do this in reverse, i.e. hard-coding the answers for 2.0
from astropy import __version__ as astropy_version  # nopep8
if astropy_version >= '2.0':
    from astropy import constants
    from astropy.constants import (si, astropyconst13)
    units.M_sun._represents = units.Unit(astropyconst13.M_sun)
    constants.M_sun = si.M_sun = astropyconst13.M_sun
    constants.G = si.G = astropyconst13.G
    constants.c = si.c = astropyconst13.c
    constants.pc = si.pc = astropyconst13.pc

# something changed in scipy 0.19, something FFT-related
if scipy_version < '0.19':
    TEST_RESULTS = {
        'inspiral_range': 19.63704209223392 * units.Mpc,
        'inspiral_range_psd': 7.915847068684727 * units.Mpc ** 2 / units.Hz,
        'burst_range': 13.813232309724613 * units.Mpc,
        'burst_range_spectrum': 35.19303454822539 * units.Mpc,
    }
else:
    TEST_RESULTS = {
        'inspiral_range': 19.63872448570372 * units.Mpc,
        'inspiral_range_psd': 7.92640311063505 * units.Mpc ** 2 / units.Hz,
        'burst_range': 13.815456279746522 * units.Mpc,
        'burst_range_spectrum': 35.216492263916535 * units.Mpc,
    }


# -- utilities ----------------------------------------------------------------

@pytest.fixture(scope='module')
def psd():
    h5path = os.path.join(utils.TEST_DATA_DIR, 'HLV-HW100916-968654552-1.hdf')
    try:
        data = TimeSeries.read(h5path, 'L1:LDAS-STRAIN', format='hdf5')
    except ImportError as e:
        pytest.skip(str(e))
    return data.psd(.4, overlap=.2, window=('kaiser', 24))


# -- gwpy.astro.range ---------------------------------------------------------

def test_inspiral_range_psd(psd):
    """Test for :func:`gwpy.astro.inspiral_range_psd`
    """
    r = astro.inspiral_range_psd(psd[1:])  # avoid DC
    assert isinstance(r, FrequencySeries)
    utils.assert_quantity_almost_equal(r.max(),
                                       TEST_RESULTS['inspiral_range_psd'])


def test_inspiral_range(psd):
    """Test for :func:`gwpy.astro.inspiral_range_psd`
    """
    r = astro.inspiral_range(psd, fmin=40)
    utils.assert_quantity_almost_equal(r, TEST_RESULTS['inspiral_range'])


def test_burst_range(psd):
    """Test for :func:`gwpy.astro.burst_range`
    """
    r = astro.burst_range(psd.crop(None, 1000)[1:])
    utils.assert_quantity_almost_equal(r, TEST_RESULTS['burst_range'])


def test_burst_range_spectrum(psd):
    """Test for :func:`gwpy.astro.burst_range_spectrum`
    """
    r = astro.burst_range_spectrum(psd.crop(None, 1000)[1:])
    utils.assert_quantity_almost_equal(r.max(),
                                       TEST_RESULTS['burst_range_spectrum'])



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""LIGO_LW I/O registrations for gwpy.frequencyseries objects
"""

from ...io import registry as io_registry
from ...io.ligolw import is_ligolw
from ...types.io.ligolw import read_series
from .. import FrequencySeries

# -- registration -------------------------------------------------------------

io_registry.register_reader('ligolw', FrequencySeries, read_series)
io_registry.register_identifier('ligolw', FrequencySeries, is_ligolw)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for plotting data with matplotlib and LAL
"""

import itertools
import re

from matplotlib import rcParams

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

rUNDERSCORE = re.compile(r'(?<!\\)_(?!.*{)')

# groups of input parameters (for passing to Plot() and subclasses)
FIGURE_PARAMS = [
    'figsize',
]
AXES_PARAMS = [
    'projection', 'sharex', 'sharey', 'xlim', 'ylim', 'xlabel', 'ylabel',
    'xscale', 'yscale', 'title', 'epoch',
]
LINE_PARAMS = [
    'linewidth', 'linestyle', 'color', 'label', 'alpha', 'rasterized',
]
COLLECTION_PARAMS = [
    'cmap', 'vmin', 'vmax', 'marker', 's', 'norm', 'rasterized',
]
IMAGE_PARAMS = [
    'imshow', 'cmap', 'vmin', 'vmax', 'norm', 'rasterized', 'extent',
    'origin', 'interpolation', 'aspect',
]
HIST_PARAMS = [
    'bins', 'range', 'normed', 'weights', 'cumulative', 'bottom',
    'histtype', 'align', 'orientation', 'rwidth', 'log', 'color',
    'label', 'stacked', 'logbins',
]
LEGEND_PARAMS = [
    'loc', 'borderaxespad', 'ncol',
]
ARTIST_PARAMS = set(itertools.chain.from_iterable([
    LINE_PARAMS,
    COLLECTION_PARAMS,
    IMAGE_PARAMS,
    HIST_PARAMS,
]))


def color_cycle(colors=None):
    """An infinite iterator of the given (or default) colors
    """
    if colors:
        return itertools.cycle(colors)
    return itertools.cycle(rcParams['axes.color_cycle'])


def marker_cycle(markers=None):
    """An infinite iterator of the given (or default) markers
    """
    if markers:
        return itertools.cycle(markers)
    return itertools.cycle(('o', 'x', '+', '^', 'D', 'H', '1'))



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.kerberos`
"""

import os
import sys

import pytest

from ...tests.mocks import mock
from ...tests.utils import TemporaryFilename
from .. import kerberos as io_kerberos

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# remove real user's keytab, if present
_KTNAME = os.environ.get('KRB5_KTNAME', '_NO KT_')

# mocked klist output
KLIST = b"""Keytab name: FILE:/test.keytab
KVNO Principal
---- -------------------------------
   1 albert.einstein@LIGO.ORG
   2 ronald.drever@LIGO.ORG"""


def setup_module():
    os.environ.pop('KRB5_KTNAME', None)


def teardown_module():
    if _KTNAME != '_NO KT_':
        os.environ['KRB5_KTNAME'] = _KTNAME


def mock_popen_return(popen, out='', err='', returncode=0):
    mocked_p = mock.Mock()
    mocked_p.__enter__ = mock.Mock(return_value=mocked_p)
    mocked_p.__exit__ = mock.Mock(return_value=None)
    mocked_p.configure_mock(**{
        'communicate.return_value': (out, err),
        'poll.return_value': returncode,
        'returncode': returncode,
    })
    popen.return_value = mocked_p


@mock.patch('subprocess.Popen')
def test_parse_keytab(mocked_popen):
    mock_popen_return(mocked_popen, out=KLIST)

    # assert principals get extracted correctly
    principals = io_kerberos.parse_keytab('test.keytab')
    assert principals == [['albert.einstein', 'LIGO.ORG'],
                          ['ronald.drever', 'LIGO.ORG']]

    # assert klist fail gets raise appropriately
    mock_popen_return(mocked_popen, returncode=1)
    with pytest.raises(io_kerberos.KerberosError):
        io_kerberos.parse_keytab('test.keytab')


@mock.patch('gwpy.io.kerberos.which', return_value='/bin/kinit')
@mock.patch('subprocess.Popen')
@mock.patch('getpass.getpass', return_value='test')
@mock.patch('gwpy.io.kerberos.input', return_value='rainer.weiss')
def test_kinit(raw_input_, getpass, mocked_popen, which, capsys):
    mocked_popen.return_value.poll.return_value = 0

    # default popen kwargs
    popen_kwargs = {'stdin': -1, 'stdout': -1, 'env': None}

    # pass username and password, and kinit exe path
    io_kerberos.kinit(username='albert.einstein', password='test',
                      exe='/usr/bin/kinit', verbose=True)
    mocked_popen.assert_called_with(
        ['/usr/bin/kinit', 'albert.einstein@LIGO.ORG'], **popen_kwargs)
    out, err = capsys.readouterr()
    assert out == (
        'Kerberos ticket generated for albert.einstein@LIGO.ORG\n')

    # configure klisting (remove Drever)
    mock_popen_return(mocked_popen, out=KLIST.rsplit(b'\n', 1)[0])
    os.environ['KRB5_KTNAME'] = '/test.keytab'

    # test keytab from environment not found (default) prompts user
    io_kerberos.kinit()
    mocked_popen.assert_called_with(
        ['/bin/kinit', 'rainer.weiss@LIGO.ORG'], **popen_kwargs)

    # test keytab from enviroment found
    with TemporaryFilename(suffix='.keytab') as tmp:
        io_kerberos.kinit(keytab=tmp)
        mocked_popen.assert_called_with(
            ['/bin/kinit', '-k', '-t', tmp, 'albert.einstein@LIGO.ORG'],
            **popen_kwargs)

    os.environ.pop('KRB5_KTNAME', None)

    # pass keytab
    io_kerberos.kinit(keytab='test.keytab')
    mocked_popen.assert_called_with(
        ['/bin/kinit', '-k', '-t', 'test.keytab',
         'albert.einstein@LIGO.ORG'], **popen_kwargs)

    # don't pass keytab (prompts for username and password)
    io_kerberos.kinit()
    getpass.assert_called_with(
        prompt='Password for rainer.weiss@LIGO.ORG: ', stream=sys.stdout)
    mocked_popen.assert_called_with(
        ['/bin/kinit', 'rainer.weiss@LIGO.ORG'], **popen_kwargs)

    # test using krb5ccname (credentials cache)
    io_kerberos.kinit(krb5ccname='/test_cc.krb5')
    popen_kwargs['env'] = {'KRB5CCNAME': '/test_cc.krb5'}
    mocked_popen.assert_called_with(
        ['/bin/kinit', 'rainer.weiss@LIGO.ORG'], **popen_kwargs)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/Output utilities for LAL Cache files.
"""

from __future__ import division

import os.path
import tempfile
import warnings
from collections import OrderedDict
from gzip import GzipFile

from six import string_types
from six.moves import StringIO

try:
    from lal.utils import CacheEntry
except ImportError:
    HAS_CACHEENTRY = False
else:
    HAS_CACHEENTRY = True

try:
    from glue.lal import Cache
except ImportError:
    HAS_CACHE = False
else:
    HAS_CACHE = True

from ..time import LIGOTimeGPS

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# build list of file-like types
try:  # python2.x
    FILE_LIKE = (
        file, GzipFile,
        tempfile._TemporaryFileWrapper,  # pylint: disable=protected-access
    )
except NameError:  # python3.x
    from io import IOBase
    FILE_LIKE = (
        IOBase, GzipFile,
        tempfile._TemporaryFileWrapper,  # pylint: disable=protected-access
    )


# -- cache I/O ----------------------------------------------------------------

def read_cache(lcf, coltype=LIGOTimeGPS):
    """Read a LAL-format cache file

    Parameters
    ----------
    lcf : `str`, `file`
        Input file or file path to read

    coltype : `LIGOTimeGPS`, `int`, optional
        Type for GPS times

    Returns
    -------
    cache : `list` of :class:`lal.utils.CacheEntry`

    Notes
    -----
    This method requires |lal|_.
    """
    from lal.utils import CacheEntry  # pylint: disable=redefined-outer-name

    # open file
    if not isinstance(lcf, FILE_LIKE):
        with open(lcf, 'r') as fobj:
            return read_cache(fobj, coltype=coltype)

    # read file
    out = []
    append = out.append
    for line in lcf:
        if isinstance(line, bytes):
            line = line.decode('utf-8')
        append(CacheEntry(line, coltype=coltype))
    return out


def open_cache(*args, **kwargs):  # pragma: no cover
    # pylint: disable=missing-docstring
    warnings.warn("gwpy.io.cache.open_cache was renamed read_cache",
                  DeprecationWarning)
    return read_cache(*args, **kwargs)


def write_cache(cache, fobj):
    """Write a `list` of cache entries to a file

    Parameters
    ----------
    cache : `list` of :class:`lal.utils.CacheEntry`
        The cache to write

    fobj : `file`, `str`
        The open file object, or file path to write to.
    """
    # open file
    if isinstance(fobj, string_types):
        with open(fobj, 'w') as fobj2:
            return write_cache(cache, fobj2)

    # write file
    for entry in cache:
        line = '{0}\n'.format(entry)
        try:
            fobj.write(line)
        except TypeError:  # python3 'wb' mode
            fobj.write(line.encode('utf-8'))


def is_cache(cache):
    """Returns `True` if ``cache`` is a readable cache file or object

    Parameters
    ----------
    cache : `str`, `file`, `list`
        Object to detect as cache

    Returns
    -------
    iscache : `bool`
        `True` if the input object is a cache, or a file in LAL cache format,
        otherwise `False`
    """
    if isinstance(cache, string_types + FILE_LIKE):
        try:
            return bool(len(read_cache(cache)))
        except (TypeError, ValueError, UnicodeDecodeError, ImportError):
            # failed to parse cache
            return False
    if HAS_CACHE and isinstance(cache, Cache):
        return True
    if (isinstance(cache, (list, tuple)) and cache and
            all(map(is_cache_entry, cache))):
        return True

    return False


def is_cache_entry(path):
    """Returns `True` if ``path`` can be represented as a cache entry

    In practice this just tests whether the input is |LIGO-T050017|_ compliant.

    Parameters
    ----------
    path : `str`, :class:`lal.utils.CacheEntry`
        The input to test

    Returns
    -------
    isentry : `bool`
        `True` if ``path`` is an instance of `CacheEntry`, or can be parsed
        using |LIGO-T050017|_.
    """
    if HAS_CACHEENTRY and isinstance(path, CacheEntry):
        return True
    try:
        file_segment(path)
    except (ValueError, TypeError, AttributeError):
        return False
    return True


# -- cache manipulation -------------------------------------------------------

def file_list(flist):
    """Parse a number of possible input types into a list of filepaths.

    Parameters
    ----------
    flist : `file-like` or `list-like` iterable
        the input data container, normally just a single file path, or a list
        of paths, but can generally be any of the following

        - `str` representing a single file path (or comma-separated collection)
        - open `file` or `~gzip.GzipFile` object
        - :class:`~lal.utils.CacheEntry`
        - `str` with ``.cache`` or ``.lcf`` extension
        - simple `list` or `tuple` of `str` paths

    Returns
    -------
    files : `list`
        `list` of `str` file paths

    Raises
    ------
    ValueError
        if the input `flist` cannot be interpreted as any of the above inputs
    """
    # open a cache file and return list of paths
    if isinstance(flist, string_types) and flist.endswith(('.cache', '.lcf')):
        return [e.path for e in read_cache(flist)]

    # separate comma-separate list of names
    if isinstance(flist, string_types):
        return flist.split(',')

    # parse list of entries (of some format)
    if isinstance(flist, (list, tuple)):
        return list(map(file_name, flist))

    # otherwise parse a single entry
    try:
        return [file_name(flist)]
    except ValueError as exc:
        exc.args = ("Could not parse input %r as one or more "
                    "file-like objects" % flist,)
        raise


def file_name(fobj):
    """Returns the name (path) of the file object
    """
    if isinstance(fobj, string_types):
        return fobj
    if isinstance(fobj, FILE_LIKE) and not isinstance(fobj, StringIO):
        return fobj.name
    if HAS_CACHEENTRY and isinstance(fobj, CacheEntry):
        return fobj.path
    raise ValueError("Cannot parse file name for %r" % fobj)


def file_segment(filename):
    """Return the data segment for a filename following T050017

    Parameters
    ---------
    filename : `str`, :class:`~lal.utils.CacheEntry`
        the path name of a file

    Returns
    -------
    segment : `~gwpy.segments.Segment`
        the ``[start, stop)`` GPS segment covered by the given file

    Notes
    -----
    |LIGO-T050017|_ declares a filenaming convention that includes
    documenting the GPS start integer and integer duration of a file,
    see that document for more details.
    """
    from ..segments import Segment
    try:  # CacheEntry
        return Segment(filename.segment)
    except AttributeError:  # file path (str)
        base = os.path.basename(filename)
        try:
            _, _, start, end = base.split('-')
        except ValueError as exc:
            exc.args = ('Failed to parse {0!r} as LIGO-T050017-compatible '
                        'filename'.format(base),)
            raise
        start = float(start)
        end = float(end.split('.')[0])
        return Segment(start, start+end)


def cache_segments(*caches):
    """Returns the segments of data covered by entries in the cache(s).

    Parameters
    ----------
    *caches : `list`
        One or more lists of file paths
        (`str` or :class:`~lal.utils.CacheEntry`).

    Returns
    -------
    segments : `~gwpy.segments.SegmentList`
        A list of segments for when data should be available
    """
    from ..segments import SegmentList
    out = SegmentList()
    for cache in caches:
        out.extend(file_segment(e) for e in cache)
    return out.coalesce()


def flatten(*caches):
    """Flatten a nested list of cache entries

    Parameters
    ----------
    *caches : `list`
        One or more lists of file paths
        (`str` or :class:`~lal.utils.CacheEntry`).

    Returns
    -------
    flat : `list`
        A flat `list` containing the unique set of entries across
        each input.
    """
    return list(OrderedDict.fromkeys(e for c in caches for e in c))


def find_contiguous(*caches):
    """Separate one or more cache entry lists into time-contiguous sub-lists

    Parameters
    ----------
    *caches : `list`
        One or more lists of file paths
        (`str` or :class:`~lal.utils.CacheEntry`).

    Returns
    -------
    caches : `iter` of `list`
        an interable yielding each contiguous cache
    """
    flat = flatten(*caches)
    for segment in cache_segments(flat):
        yield sieve(flat, segment=segment)


def sieve(cache, segment=None):
    """Filter the cache to find those entries that overlap ``segment``

    Parameters
    ----------
    cache : `list`
        Input list of file paths

    segment : `~gwpy.segments.Segment`
        The ``[start, stop)`` interval to match against.
    """
    return type(cache)(e for e in cache if segment.intersects(file_segment(e)))



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting `EventTable` rate versus time for specific column bins

I would like to study the rate at which event triggers are generated by the
`ExcessPower` gravitational-wave burst detection algorithm, over a small
stretch of data, binned by various thresholds on signal-to-noise ratio (SNR).

The data from which these events were generated contain a simulated
gravitational-wave signal, or hardware injection, used to validate
the performance of the LIGO detectors and downstream data analysis procedures.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.table'

# First, we import the `EventTable` object and read in a set of events from
# a LIGO_LW-format XML file containing a
# :class:`sngl_burst <glue.ligolw.lsctables.SnglBurstTable>` table
from gwpy.table import EventTable
events = EventTable.read('H1-LDAS_STRAIN-968654552-10.xml.gz',
                         tablename='sngl_burst', columns=['peak', 'snr'])

# .. note::
#
#    Here we manually specify the `columns` to read in order to optimise
#    the `read()` operation to parse only the data we actually need.

# Now we can use the :meth:`~EventTable.binned_event_rates` method to
# calculate the event rate in a number of bins of SNR.
rates = events.binned_event_rates(1, 'snr', [2, 3, 5, 8], operator='>=',
                                  start=968654552, end=968654562)
# .. note::
#
#    The list `[2, 3, 5, 8]` and operator `>=` specifies SNR tresholds of
#    2, 3, 5, and 8.

# Finally, we can make a plot:
plot = rates.step()
ax = plot.gca()
ax.set_xlim(968654552, 968654562)
ax.set_ylabel('Event rate [Hz]')
ax.set_title('LIGO Hanford Observatory event rate for HW100916')
ax.legend(title='SNR $>=$')
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""I/O utilities for LIGO Open Science Center data

For more details, see https://losc.ligo.org
"""

import json
from collections import namedtuple
from operator import attrgetter

from astropy.utils.data import get_readable_fileobj

from ..segments import (Segment, SegmentList)
from ..time import to_gps

# default URL
LOSC_URL = 'https://losc.ligo.org'


# -- JSON handling ------------------------------------------------------------

def fetch_json(url, verbose=False):
    """Fetch JSON data from a remote URL

    Parameters
    ----------
    url : `str`
        the remote URL to fetch

    verbose : `bool`, optional
        display verbose download progress, default: `False`

    Returns
    ------
    json : `object`
        the data fetched from ``url`` as parsed by :func:`json.loads`

    See also
    --------
    json.loads
        for details of the JSON parsing

    Examples
    --------
    >>> from gwpy.io.losc import fetch_json
    >>> fetch_json('https://losc.ligo.org/archive/1126257414/1126261510/json/')
    """
    with get_readable_fileobj(url, show_progress=verbose) as response:
        data = response.read()
        try:
            return json.loads(data)
        except ValueError as exc:
            exc.args = ("Failed to parse LOSC JSON from %r: %s"
                        % (url, str(exc)),)
            raise


# -- API calls ----------------------------------------------------------------

def fetch_dataset_json(gpsstart, gpsend, host=LOSC_URL):
    """Returns the JSON metadata for all datasets matching the GPS interval

    Parameters
    ----------
    gpsstart : `int`
        the GPS start of the desired interval

    gpsend : `int`
        the GPS end of the desired interval

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    json
        the JSON data retrieved from LOSC and returned by `json.loads`
    """
    url = '{}/archive/{:d}/{:d}/json/'.format(host, gpsstart, gpsend)
    return fetch_json(url)


def fetch_event_json(event, host=LOSC_URL):
    """Returns the JSON metadata for the given event
    """
    url = '{}/archive/{}/json/'.format(host, event)
    return fetch_json(url)


def fetch_run_json(run, detector, gpsstart, gpsend, host=LOSC_URL):
    """Returns the JSON metadata for the given science run parameters

    Parameters
    ----------
    run : `str`
        the name of the science run, e.g. ``'O1'``

    detector : `str`
        the prefix of the GW detector, e.g. ``'L1'``

    gpsstart : `int`
        the GPS start of the desired interval

    gpsend : `int`
        the GPS end of the desired interval

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    json
        the JSON data retrieved from LOSC and returned by `json.loads`
    """
    url = '{}/archive/links/{}/{}/{:d}/{:d}/json/'.format(
        host, run, detector, gpsstart, gpsend)
    return fetch_json(url)


# -- utilities ----------------------------------------------------------------

def event_gps(event, host=LOSC_URL):
    """Returns the GPS time of an open-data event

    Parameters
    ----------
    event : `str`
        the name of the event to query

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    gps : `float`
        the GPS time of this event
    """
    return fetch_event_json(event, host=host)['GPS']


def event_segment(event, host=LOSC_URL, **match):
    """Returns the GPS segment covered by a LOSC event dataset

    Parameters
    ----------
    event : `str`
        the name of the event to query

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    **match : metadata requirements, optional
        restrictions for matched files, e.g. ``detector='L1'``, all
        keys must be present in JSON metadata for all files

    Returns
    -------
    segment : `~gwpy.segments.Segment`
        the GPS ``[start, end)`` segment matched for this event

    Examples
    --------
    >>> from gwpy.io.losc import event_segment
    >>> event_segment('GW150914')
    [1126257414 ... 1126261510)
    """
    jdata = fetch_event_json(event, host=host)
    seg = None
    for fmeta in sieve_urls(jdata['strain'], **match):
        start = fmeta['GPSstart']
        end = start + fmeta['duration']
        fseg = Segment(start, end)
        if seg is None:
            seg = fseg
        else:
            seg |= fseg
    if seg is None:
        raise ValueError("No files matched for event {}".format(event))
    return seg


def sieve_urls(urllist, **match):
    """Sieve a list of LOSC URL metadata dicts based on key, value pairs

    This method simply matches keys from the ``match`` keywords with those
    found in the JSON dicts for a file URL returned by the LOSC API.
    """
    for urlmeta in urllist:
        if any(match[key] != urlmeta[key] for key in match):
            continue
        yield urlmeta


# -- epoch discovery ----------------------------------------------------------

def find_datasets(start, end, detector=None, strict=False, host=LOSC_URL):
    """Returns the datasets that could contain data for the given interval

    Parameters
    ----------
    start : `int`
        the GPS start time of the interval

    end : `int`
        the GPS end time of the interval

    detector : `str`, optional
        the observatory prefix for the detector of interest, e.g. ``'L1'``

    strict : `bool`, optional
        return only those datasets with 100% coverage of the interval,
        default: `False` (return any dataset overlapping the interval)

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    datasets : `tuple`
        a `tuple` containing the name of each dataset that covers the
        requested interval

    Examples
    --------
    >>> from gwpy.io.losc import find_datasets
    >>> find_datasets(934000000, 934100000)
    (u'S6', u'tenyear')
    >>> find_datasets(1126257414, 1126261510)
    (u'GW150914', u'tenyear')
    """
    start = to_gps(start).gpsSeconds
    end = to_gps(end).gpsSeconds
    span = Segment(start, end)
    jdata = fetch_dataset_json(start, end, host=host)

    Dataset = namedtuple('Dataset', ('name', 'span', 'gap'))

    # extract epochs
    epochs = []
    for epochtype in jdata:
        for epoch, metadata in jdata[epochtype].items():
            if detector and detector not in metadata['detectors']:
                continue

            # get dataset segment
            if epochtype == 'events' and detector:
                epochseg = event_segment(epoch, detector=detector, host=host)
            elif epochtype == 'events':
                epochseg = event_segment(epoch, host=host)
            else:
                epochseg = Segment(metadata['GPSstart'], metadata['GPSend'])

            # match segment against request
            try:
                coverage = span & epochseg
            except (ValueError, TypeError):
                continue
            gap = abs(SegmentList([span]) - SegmentList([epochseg]))
            if strict and gap:
                continue
            epochs.append(Dataset(epoch, abs(epochseg), gap))

    # sort epochs by coverage
    return list(zip(*sorted(epochs, key=attrgetter('gap', 'span', 'name'))))[0]



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The `gwpy.signal` module provides signal-processing utilities that
extend the functionality of `scipy.signal` (and others) for specific GW data
applications.

See the sub-modules for more details.
"""

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2018)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""An extension of the Plot and Axes classes for handling Series
"""

import numpy

from matplotlib import (colors, __version__ as mpl_version)
from matplotlib.projections import register_projection

from . import text
from .core import Plot
from .axes import Axes
from .decorators import auto_refresh
from ..segments import SegmentList

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = ['SeriesPlot', 'SeriesAxes']

USE_IMSHOW = mpl_version >= '2.0'  # imshow respects log scaling in >=2.0


class SeriesAxes(Axes):
    """Custom `Axes` for a `~gwpy.plotter.SeriesPlot`.
    """
    name = 'series'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('xmargin', 0)
        super(SeriesAxes, self).__init__(*args, **kwargs)

    @auto_refresh
    def plot(self, *args, **kwargs):
        """Plot data onto these Axes.

        Parameters
        ----------
        args
            a single `~gwpy.types.Series` (or sub-class)
            or standard (x, y) data arrays

        kwargs
            keyword arguments applicable to :meth:`~matplotib.axes.Axes.plot`

        Returns
        -------
        Line2D
            the `~matplotlib.lines.Line2D` for this line layer

        See Also
        --------
        matplotlib.axes.Axes.plot
            for a full description of acceptable ``*args`` and ``**kwargs``
        """
        from ..types import Series

        # plot series
        if len(args) == 1 and isinstance(args[0], Series):
            return self.plot_series(*args, **kwargs)

        # plot everything else
        return super(SeriesAxes, self).plot(*args, **kwargs)

    @auto_refresh
    def plot_series(self, series, **kwargs):
        """Plot a `~gwpy.types.Series` onto these axes

        Parameters
        ----------
        series : `~gwpy.types.Series`
            data to plot

        **kwargs
            any other keyword arguments acceptable for
            :meth:`~matplotlib.Axes.plot`

        Returns
        -------
        line : `~matplotlib.lines.Line2D`
            the newly added line

        See Also
        --------
        matplotlib.axes.Axes.plot
            for a full description of acceptable ``*args`` and ``**kwargs``
        """
        kwargs.setdefault('label', text.to_string(series.name))
        line = self.plot(series.xindex.value, series.value, **kwargs)

        # update datalim to include end of final sample
        if not kwargs.get('linestyle', '-').startswith('steps-'):
            self.update_datalim(list(zip(series.xspan, (0, 0))), updatey=False)
            self.autoscale_view()

        # set default limits (protecting against 0-log)
        if len(self._get_artists()) == 1:
            self._set_lim_from_array(series, 'x')

        # set labels from units
        if not self.get_xlabel():
            self.set_xlabel(text.unit_as_label(series.xunit))
        if not self.get_ylabel():
            self.set_ylabel(text.unit_as_label(series.unit))

        return line

    @auto_refresh
    def plot_mmm(self, mean_, min_=None, max_=None, alpha=.1, **kwargs):
        """Plot a `Series` onto these axes, with shaded regions

        The ``mean_`` `Series` is plotted normally, while the ``min_``
        and ``max_`` `Series` are plotted lightly below and above,
        with a fill between them and the ``mean_``.

        Parameters
        ----------
        mean_ : `~gwpy.types.Series`
            data to plot normally

        min_ : `~gwpy.types.Series`
            first data set to shade to ``mean_``

        max_ : `~gwpy.types.Series`
            second data set to shade to ``mean_``

        alpha : `float`, `None`, optional
            value for alpha channel for shading, default: ``0.1``

        **kwargs
            any other keyword arguments acceptable for
            :meth:`~matplotlib.Axes.plot`, only used to draw the
            lines

        Returns
        -------
        artists : `tuple`
            a 5-tuple containing:

            - `~matplotlib.lines.Line2d` for ``mean_``,
            - `~matplotlib.lines.Line2D` for ``min_``,
            - `~matplotlib.collections.PolyCollection` for ``min_`` shading,
            - `~matplotlib.lines.Line2D` for ``max_``, and
            - `~matplitlib.collections.PolyCollection` for ``max_`` shading

        See Also
        --------
        matplotlib.axes.Axes.plot
            for a full description of acceptable ``*args`` and ``**kwargs``
        """
        # plot mean
        meanline = self.plot_series(mean_, **kwargs)[0]

        # modify keywords for shading
        kwargs.update({
            'label': '',
        })
        color = kwargs.pop('color', meanline.get_color())
        linewidth = kwargs.pop('linewidth', meanline.get_linewidth()) / 2

        def _plot_shade(series):
            line, = self.plot(series, color=color, linewidth=linewidth,
                              **kwargs)
            coll = self.fill_between(series.xindex.value, series.value,
                                     mean_.value, alpha=alpha, color=color,
                                     rasterized=kwargs.get('rasterized'))
            return line, coll

        # plot lower shade
        if min_ is not None:
            minline, mincoll = _plot_shade(min_)
        else:
            minline = mincoll = None

        # plot upper shade
        if max_ is not None:
            maxline, maxcoll = _plot_shade(max_)
        else:
            maxline = maxcoll = None

        return meanline, minline, mincoll, maxline, maxcoll

    def plot_array2d(self, array, imshow=USE_IMSHOW, **kwargs):
        """Plot a 2D array onto these axes

        Parameters
        ----------
        array : `~gwpy.types.Array2D`,
            Data to plot (e.g. a `~gwpy.spectrogram.Spectrogram`)

        imshow : `bool`, optional
            If `True`, use :meth:`~matplotlib.axes.Axes.imshow` to render
            the array as an image, otherwise use
            :meth:`~matplotlib.axes.Axes.pcolormesh`, default is `True`
            with `matplotlib >= 2.0`, otherwise `False`.

        norm : ``'log'``, `~matplotlib.colors.Normalize`
            A `~matplotlib.colors.Normalize`` instance used to scale the
            colour data, or ``'log'`` to use `LogNorm`.

        **kwargs
            Any other keyword arguments acceptable for
            :meth:`~matplotlib.Axes.imshow` (if ``imshow=True``),
            or :meth:`~matplotlib.Axes.pcolormesh` (``imshow=False``)

        Returns
        -------
        layer : `~matplotlib.collections.QuadMesh`, `~matplotlib.images.Image`
            the layer for this array

        See Also
        --------
        matplotlib.axes.Axes.imshow
        matplotlib.axes.Axes.pcolormesh
            for a full description of acceptable ``*args`` and ``**kwargs``
        """
        # allow normalisation as 'log'
        if kwargs.get('norm', None) == 'log':
            vmin = kwargs.get('vmin', None)
            vmax = kwargs.get('vmax', None)
            kwargs['norm'] = colors.LogNorm(vmin=vmin, vmax=vmax)

        # plot with imshow
        if imshow:
            layer = self._imshow_array2d(array, **kwargs)

        # plot with pcolormesh
        else:
            layer = self._pcolormesh_array2d(array, **kwargs)

        # format axes
        if not self.get_ylabel():
            self.set_ylabel(text.unit_as_label(array.yunit))
        if len(self._get_artists()) == 1:  # first plotted element
            self._set_lim_from_array(array, 'x')
            self._set_lim_from_array(array, 'y')

        return layer

    def _imshow_array2d(self, array, origin='lower', interpolation='none',
                        aspect='auto', **kwargs):
        # calculate extent
        extent = tuple(array.xspan) + tuple(array.yspan)
        if self.get_xscale() == 'log' and extent[0] == 0.:
            extent = (1e-300,) + extent[1:]
        if self.get_yscale() == 'log' and extent[2] == 0.:
            extent = extent[:2] + (1e-300,) + extent[3:]
        kwargs.setdefault('extent', extent)

        return self.imshow(array.value.T, **kwargs)

    def _pcolormesh_array2d(self, array, **kwargs):
        x = numpy.concatenate((array.xindex.value, [array.xspan[-1]]))
        y = numpy.concatenate((array.yindex.value, [array.yspan[-1]]))
        xcoord, ycoord = numpy.meshgrid(x, y, copy=False, sparse=True)
        return self.pcolormesh(xcoord, ycoord, array.value.T, **kwargs)

    def _set_lim_from_array(self, array, axis):
        """Set the axis limits using the index of an `~gwpy.types.Array`
        """
        # get limits from array span
        span = getattr(array, '{}span'.format(axis))
        scale = getattr(self, 'get_{}scale'.format(axis))()
        if scale == 'log' and not span[0]:  # protect log(0)
            index = getattr(array, '{}index'.format(axis)).value
            span = index[1], span[1]

        # set limits
        set_lim = getattr(self, 'set_{}lim'.format(axis))
        return set_lim(*span)


register_projection(SeriesAxes)


class SeriesPlot(Plot):
    """`Figure` for displaying a `~gwpy.types.Series`.

    Parameters
    ----------
    *series : `Series`
        any number of `~gwpy.types.Series` to display on the plot

    **kwargs
        other keyword arguments as applicable for the
        `~gwpy.plotter.Plot`
    """
    _DefaultAxesClass = SeriesAxes

    def __init__(self, *series, **kwargs):
        """Initialise a new SeriesPlot
        """
        kwargs.setdefault('projection', self._DefaultAxesClass.name)
        self._update_kwargs_from_data(kwargs, series)

        # extract custom keyword arguments
        sep = kwargs.pop('sep', False)

        # separate keyword arguments
        axargs, plotargs = self._parse_kwargs(kwargs)
        sharex = axargs.get('sharex', False)

        # generate figure
        super(SeriesPlot, self).__init__(**kwargs)

        # plot data
        self._init_axes(series, sep, axargs, plotargs)

        # set epoch
        for ax in self.axes[:-1]:
            if sharex:
                ax.set_xlabel("")
            if 'label' in kwargs:
                ax.legend()

    def _update_kwargs_from_data(self, kwargs, data):
        """Add `__init__` keywords based on input data
        """
        flat = self._get_axes_data(data, flat=True)
        if not flat:  # bail out with no data
            return kwargs

        # set xlim based on xspans
        span = SegmentList([s.xspan for s in flat]).extent()
        if kwargs.get('xscale', 'linear') == 'log':
            span = span[0] + min(s.dx.value for s in flat), span[1]
        kwargs.setdefault('xlim', span)

    def _init_axes(self, data, sep, axargs, plotargs):
        """Initalise these Axes from the input data and keywords
        """
        sharex = axargs.pop('sharex', False)
        sharey = axargs.pop('sharey', False)
        groups = self._get_axes_data(data, sep=sep)
        for group in groups:
            ax = self._add_new_axes(**axargs)
            for series in group:
                ax.plot(series, **plotargs)
            if sharex is True:
                axargs.setdefault('sharex', ax)
            if sharey is True:
                axargs.setdefault('sharey', ax)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.datafind`
"""

from __future__ import print_function

import os

import pytest

from ...tests.utils import (skip_missing_dependency, TEST_DATA_DIR)
from ...tests import mocks
from ...tests.mocks import mock
from .. import datafind as io_datafind

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

TEST_GWF_FILE = os.path.join(TEST_DATA_DIR, 'HLV-HW100916-968654552-1.gwf')


@pytest.fixture(scope='class')
@skip_missing_dependency('lal')
def connection():
    return mocks.mock_datafind_connection(TEST_GWF_FILE)


def test_on_tape():
    assert io_datafind.on_tape(TEST_GWF_FILE) is False


def test_connect(connection):
    with mock.patch('glue.datafind.GWDataFindHTTPConnection',
                    connection), \
            mock.patch('glue.datafind.GWDataFindHTTPSConnection',
                       connection), \
            mock.patch('glue.datafind.find_credential',
                       mocks.mock_find_credential):
        io_datafind.connect()  # HTTP
        io_datafind.connect('host', 443)  # HTTPS


def test_find_frametype(connection):
    with mock.patch('glue.datafind.GWDataFindHTTPConnection') as \
            mock_connection, \
            mock.patch('gwpy.io.datafind.num_channels', lambda x: 1), \
            mock.patch('gwpy.io.datafind.iter_channel_names',
                       lambda x: ['L1:LDAS-STRAIN']):
        mock_connection.return_value = connection
        assert io_datafind.find_frametype('L1:LDAS-STRAIN',
                                          allow_tape=True) == 'HW100916'
        assert io_datafind.find_frametype('L1:LDAS-STRAIN',
                                          return_all=True) == ['HW100916']

        # test missing channel raises sensible error
        with pytest.raises(ValueError) as exc:
            io_datafind.find_frametype('X1:TEST', allow_tape=True)
        assert str(exc.value) == (
            'Cannot locate the following channel(s) '
            'in any known frametype:\n    X1:TEST')

        # test malformed channel name raises sensible error
        with pytest.raises(ValueError) as exc:
            io_datafind.find_frametype('bad channel name')
        assert str(exc.value) == ('Cannot parse interferometer prefix '
                                  'from channel name \'bad channel name\','
                                  ' cannot proceed with find()')

        # test trend sorting ends up with an error
        with pytest.raises(ValueError) as exc:
            io_datafind.find_frametype('X1:TEST.rms,s-trend',
                                       allow_tape=True)
        with pytest.raises(ValueError):
            io_datafind.find_frametype('X1:TEST.rms,m-trend',
                                       allow_tape=True)


def test_find_best_frametype(connection):
    with mock.patch('glue.datafind.GWDataFindHTTPConnection') as \
            mock_connection, \
            mock.patch('gwpy.io.datafind.num_channels', lambda x: 1), \
            mock.patch('gwpy.io.datafind.iter_channel_names',
                       lambda x: ['L1:LDAS-STRAIN']):
        mock_connection.return_value = connection
        assert io_datafind.find_best_frametype(
            'L1:LDAS-STRAIN', 968654552, 968654553) == 'HW100916'



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2018)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""I/O routines for accessing LIGO segments through LOSC
"""

from astropy.utils.data import get_readable_fileobj

from ...io.losc import (LOSC_URL, find_datasets)
from ...segments import (Segment, SegmentList)
from ...time import (LIGOTimeGPS, to_gps)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def timeline_url(flag, start, end, host=LOSC_URL):
    """Construct the LOSC API URL to the given flag and GPS [start, stop).

    This method is mainly designed to support the :func:`get_segments`
    function in the same module.

    Parameters
    ----------
    flag : `str`
        the name of the flag to query

    start : `int`, `str`
        the GPS start time (or parseable date string) to query

    end : `int`, `str`
        the GPS end time (or parseable date string) to query

    host : `str`, optional
        URL of LOSC host, default: ``'losc.ligo.org'``

    Returns
    -------
    url : `str`
        the LOSC URL that will return matching timeline segments

    Examples
    --------
    >>> from gwpy.segments.io.losc import timeline_url
    >>> print(timeline_url('H1_DATA', 'Jan 1 2010', 'Jan 2 2010'))
    https://losc.ligo.org/timeline/segments/S6/H1_DATA/946339215/86400/
    """
    start = to_gps(start).gpsSeconds
    end = to_gps(end).gpsSeconds
    ifo = flag.split('_', 1)[0]
    try:
        dataset = find_datasets(start, end, detector=ifo)[0]
    except IndexError as exc:
        exc.args = ('No datasets matching [{} ... {}) for '
                    'detector={}'.format(start, end, ifo),)
        raise

    return '{}/timeline/segments/{}/{}/{}/{}/'.format(
        host, dataset, flag, start, end-start)


def get_segments(flag, start, end, verbose=False, timeout=60, host=LOSC_URL):
    """Download segments for a given flag within the GPS interval.

    Parameters
    ----------
    flag : `str`
        the name of the flag to query

    start : `int`, `str`
        the GPS start time (or parseable date string) to query

    end : `int`, `str`
        the GPS end time (or parseable date string) to query

    verbose : `bool`, optional
        show verbose download progress, default: `False`

    timeout : `int`, optional
        timeout for download (seconds)

    host : `str`, optional
        URL of LOSC host, default: ``'losc.ligo.org'``

    Returns
    -------
    segments : `~gwpy.segments.SegmentList`
        the list of GPS ``[start, end)`` segments

    Examples
    --------
    To download the list of GPS segments during which LIGO-Hanford was
    operational on Jan 1 2010:

    >>> from gwpy.segments.io.losc import get_segments
    >>> print(get_segments('H1_DATA', 'Jan 1 2010', 'Jan 2 2010'))
    [[946340946 ... 946351800)
     [946356479 ... 946360620)
     [946362652 ... 946369150)
     [946372854 ... 946382630)
     [946395595 ... 946396751)
     [946400173 ... 946404977)
     [946412312 ... 946413577)
     [946415770 ... 946422986)]
    """
    span = Segment(to_gps(start), to_gps(end))
    url = timeline_url(flag, start, end, host=host)
    segments = SegmentList()
    with get_readable_fileobj(url, show_progress=verbose,
                              remote_timeout=timeout) as response:
        data = response.read()
        for line in data.splitlines():
            start, end, duration = map(int, line.split())
            if duration != end - start:
                raise RuntimeError("Corrupt segment received from LOSC: "
                                   "printed duration does not match segment "
                                   "boundaries: {}".format(line.rstrip()))
            segments.append(Segment(LIGOTimeGPS(start), LIGOTimeGPS(end)))
    return SegmentList([span]) & segments



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""DEPRECATED - this module has been replaced by :mod:`gwpy.plot`
"""

import warnings

from matplotlib import pyplot

# utilities
from .rc import DEFAULT_PARAMS as GWPY_PLOT_PARAMS
from . import (  # pylint: disable=unused-import
    gps,  # GPS timing scales and formats
    log,  # Logarithimic scaling mods
)

# figure and axes extensions
from .core import Plot
from .axes import Axes
from .timeseries import (TimeSeriesPlot, TimeSeriesAxes)
from .spectrogram import (SpectrogramPlot)
from .frequencyseries import (FrequencySeriesPlot, FrequencySeriesAxes)
from .segments import (SegmentPlot, SegmentAxes)
from .filter import (BodePlot)
from .table import (EventTablePlot, EventTableAxes)
from .histogram import (HistogramPlot, HistogramAxes)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# pyplot.figure replacement
def figure(*args, **kwargs):  # pylint: disable=missing-docstring
    kwargs.setdefault('FigureClass', Plot)
    return pyplot.figure(*args, **kwargs)
figure.__doc__ = pyplot.figure.__doc__


warnings.warn('this module was replaced by `gwpy.plot` and will be '
              'removed in a future release', DeprecationWarning)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read/write series in LIGO_LW-XML
"""

from .. import Series
from ...io.ligolw import read_ligolw
from ...time import to_gps

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def series_contenthandler():
    """Build a `~xml.sax.handlers.ContentHandler` to read a LIGO_LW <Array>
    """
    from glue.ligolw import (
        ligolw,
        array as ligolw_array,
        param as ligolw_param
    )

    @ligolw_array.use_in
    @ligolw_param.use_in
    class ArrayContentHandler(ligolw.LIGOLWContentHandler):
        """`~xml.sax.handlers.ContentHandler` to read a LIGO_LW <Array>
        """
        pass

    return ArrayContentHandler


def read_series(source, name, match=None):
    """Read a `Series` from LIGO_LW-XML

    Parameters
    ----------
    source : `file`, `str`, :class:`~glue.ligolw.ligolw.Document`
        file path or open LIGO_LW-format XML file

    name : `str`
        name of the relevant `LIGO_LW` element to read

    match : `dict`, optional
        dict of (key, value) `Param` pairs to match correct LIGO_LW element,
        this is useful if a single file contains multiple `LIGO_LW` elements
        with the same name
    """
    from glue.ligolw.ligolw import (LIGO_LW, Time, Array, Dim)
    from glue.ligolw.param import get_param

    # read document
    xmldoc = read_ligolw(source, contenthandler=series_contenthandler())

    # parse match dict
    if match is None:
        match = dict()

    def _is_match(elem):
        try:
            if elem.Name != name:
                return False
        except AttributeError:  # Name is not set
            return False
        for key, value in match.items():
            try:
                if get_param(elem, key).pcdata != value:
                    return False
            except ValueError:  # no Param with this Name
                return False
        return True

    # parse out correct element
    matches = filter(_is_match, xmldoc.getElementsByTagName(LIGO_LW.tagName))
    try:
        elem, = matches
    except ValueError as exc:
        if not matches:
            exc.args = ("no LIGO_LW elements found matching request",)
        else:
            exc.args = ('multiple LIGO_LW elements found matching request, '
                        'please consider using `match=` to select the '
                        'correct element',)
        raise

    # get data
    array, = elem.getElementsByTagName(Array.tagName)

    # parse dimensions
    dims = array.getElementsByTagName(Dim.tagName)
    xdim = dims[0]
    x0 = xdim.Start
    dx = xdim.Scale
    xunit = xdim.Unit
    try:
        ndim = dims[1].n
    except IndexError:
        pass
    else:
        if ndim > 2:
            raise ValueError("Cannot parse LIGO_LW Array with {} "
                             "dimensions".format(ndim))

    # parse metadata
    array_kw = {
        'name': array.Name,
        'unit': array.Unit,
        'xunit': xunit,
    }
    try:
        array_kw['epoch'] = to_gps(
            elem.getElementsByTagName(Time.tagName)[0].pcdata)
    except IndexError:
        pass
    for key in ('channel',):
        try:
            array_kw[key] = get_param(elem, key)
        except ValueError:
            pass

    # build Series
    try:
        xindex, value = array.array
    except ValueError:  # not two dimensions stored
        return Series(array.array[0], x0=x0, dx=dx, **array_kw)
    return Series(value, xindex=xindex, **array_kw)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""ASCII I/O registrations for gwpy.timeseries objects
"""

from ...types.io.ascii import register_ascii_series_io
from .. import (TimeSeries, StateVector)

# -- registration -------------------------------------------------------------

register_ascii_series_io(TimeSeries, format='txt')
register_ascii_series_io(TimeSeries, format='csv', delimiter=',')
register_ascii_series_io(StateVector, format='txt')
register_ascii_series_io(StateVector, format='csv', delimiter=',')



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for unified input/output
"""

import gzip

from six import string_types

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

GZIP_SIGNATURE = b'\x1f\x8b\x08'


def identify_factory(*extensions):
    """Factory function to create I/O identifiers for a set of extensions

    The returned function is designed for use in the unified I/O registry
    via the `astropy.io.registry.register_identifier` hool.

    Parameters
    ----------
    extensions : `str`
        one or more file extension strings

    Returns
    -------
    identifier : `callable`
        an identifier function that tests whether an incoming file path
        carries any of the given file extensions (using `str.endswith`)
    """
    def identify(origin, filepath, fileobj, *args, **kwargs):
        """Identify the given extensions in a file object/path
        """
        # pylint: disable=unused-argument
        if (isinstance(filepath, string_types) and
                filepath.endswith(extensions)):
            return True
        return False
    return identify


def gopen(name, *args, **kwargs):
    """Open a file handling optional gzipping

    If ``name`` endswith ``'.gz'``, or if the GZIP file signature is
    found at the beginning of the file, the file will be opened with
    `gzip.open`, otherwise a regular file will be returned from `open`.

    Parameters
    ----------
    name : `str`
        path (name) of file to open.

    *args, **kwargs
        other arguments to pass to either `open` for regular files, or
        `gzip.open` for gzipped files.

    Returns
    -------
    file : `io.TextIoBase`, `file`, `gzip.GzipFile`
        the open file object
    """
    # filename declares gzip
    if name.endswith('.gz'):
        return gzip.open(name, *args, **kwargs)

    # open regular file
    fobj = open(name, *args, **kwargs)
    sig = fobj.read(3)
    fobj.seek(0)
    if sig == GZIP_SIGNATURE:  # file signature declares gzip
        fobj.close()  # GzipFile won't close orig file when it closes
        return gzip.open(name, *args, **kwargs)
    return fobj



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Joseph Areeda (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
#

""" Q-transform plots
"""

import os.path
import re
import warnings

from astropy.units import Quantity

from ..segments import Segment
from ..time import to_gps
from .spectrogram import (FFTMixin, Spectrogram)


class Qtransform(Spectrogram):
    """Plot the Q-transform (Omega)"""
    MAX_DATASETS = 1
    action = 'qtransform'

    def __init__(self, *args, **kwargs):
        super(Qtransform, self).__init__(*args, **kwargs)

        args = self.args
        self.qxfrm_args = {
            'gps': float(args.gps),
            'search': args.search / 2.,
            'fres': 0.5,
            'tres': 0.002,
            'whiten': not args.nowhiten,
        }
        if args.qrange is not None:
            self.qxfrm_args['qrange'] = args.qrange
        if args.frange is not None:
            self.qxfrm_args['frange'] = args.frange

    @classmethod
    def init_data_options(cls, parser):
        # call super of FFTMixin to skip setting FFT arguments
        super(FFTMixin, cls).init_data_options(parser)
        cls.arg_qxform(parser)

    @classmethod
    def arg_channels(cls, parser):
        group = parser.add_argument_group('Data options', 'What data to load')
        group.add_argument('--chan', required=True, help='Channel name.')
        group.add_argument('--gps', type=to_gps, required=True,
                           help='Central time of transform')
        group.add_argument('--search', type=float, default=64,
                           help='Time window around GPS to search')
        return group

    @classmethod
    def arg_signal(cls, parser):
        group = super(Qtransform, cls).arg_signal(parser)
        group.add_argument('--sample-freq', type=float, default=2048,
                           help='Downsample freq')

    @classmethod
    def arg_plot(cls, parser):
        group = super(Qtransform, cls).arg_plot(parser)

        # remove --out option
        outopt = [act for act in group._actions if act.dest == 'out'][0]
        group._remove_action(outopt)

        # and replace with --outdir
        group.add_argument('--outdir', default=os.path.curdir, dest='out',
                           type=os.path.abspath,
                           help='Directory for output images')

        return group

    @classmethod
    def arg_qxform(cls, parser):
        """Add an `~argparse.ArgumentGroup` for Q-transform options
        """
        group = parser.add_argument_group('Q-transform options')
        group.add_argument('--plot', nargs='+', type=float, default=[.5],
                           help='One or more times to plot')
        group.add_argument('--frange', nargs=2, type=float,
                           help='Frequency range to plot')
        group.add_argument('--qrange', nargs=2, type=float,
                           help='Search Q range')

        group.add_argument('--nowhiten', action='store_true',
                           help='do not whiten input before transform')

    def _finalize_arguments(self, args):
        """Derive standard args from our weird ones
        :type args: Namespace with command line arguments
        """
        gps = args.gps
        search = args.search

        args.start = [[int(gps - search/2)]]
        if args.epoch is None:
            args.epoch = args.gps
        args.duration = search

        args.chan = [[args.chan]]

        if args.color_scale is None:
            args.color_scale = 'linear'

        args.overlap = 0  # so that FFTMixin._finalize_arguments doesn't fail

        xmin = args.xmin
        xmax = args.xmax

        super(Qtransform, self)._finalize_arguments(args)

        # unset defaults from `TimeDomainProduct`
        args.xmin = xmin
        args.xmax = xmax

    def get_ylabel(self):
        """Default text for y-axis label"""
        return 'Frequency (Hz)'

    def get_color_label(self):
        return 'Normalized energy'

    def get_suptitle(self):
        return 'Q-transform: {0}'.format(self.chan_list[0])

    def get_title(self):
        """Default title for plot
        """
        def fformat(x):  # float format
            if isinstance(x, (list, tuple)):
                return '[{0}]'.format(', '.join(map(fformat, x)))
            if isinstance(x, Quantity):
                x = x.value
            elif isinstance(x, str):
                warnings.warn('WARNING: fformat called with a' +
                            ' string. This has ' +
                            'been depricated and may disappear ' +
                            'in a future release.')
                x = float(x)
            return '{0:.2f}'.format(x)

        bits = [('Q', fformat(self.result.q))]
        if self.qxfrm_args.get('qrange'):
            bits.append(('q-range', fformat(self.qxfrm_args['qrange'])))
        if self.qxfrm_args['whiten']:
            bits.append(('whitened',))
        bits.extend([
            ('calc f-range', fformat(self.result.yspan)),
            ('calc e-range', fformat((self.result.min(), self.result.max()))),
        ])
        return ', '.join(['='.join(bit) for bit in bits])

    def get_spectrogram(self):
        args = self.args

        asd = self.timeseries[0].asd().value
        if (asd.min() == 0):
            self.log(0, 'Input data has a zero in ASD. '
                     'Q-transform not possible.')
            self.got_error = True
            qtrans = None
        else:
            gps = self.qxfrm_args['gps']
            outseg = Segment(gps, gps).protract(args.plot[self.plot_num])
            qtrans = self.timeseries[0].q_transform(
                outseg=outseg,
                **self.qxfrm_args)

            if args.ymin is None:  # set before Spectrogram.make_plot
                args.ymin = qtrans.yspan[0]

        return qtrans

    def scale_axes_from_data(self):
        self.args.xmin, self.args.xmax = self.result.xspan
        return super(Qtransform, self).scale_axes_from_data()

    def has_more_plots(self):
        """any ranges left to plot?
        """
        return self.plot_num < len(self.args.plot)

    def save(self, outdir):  # pylint: disable=arguments-differ
        cname = re.sub('[-_:]', '_', self.timeseries[0].channel.name).replace(
            '_', '-', 1)
        png = '{0}-{1}-{2}.png'.format(cname, float(self.args.gps),
                                       self.args.plot[self.plot_num])
        outfile = os.path.join(outdir, png)
        return super(Qtransform, self).save(outfile)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.utils`
"""

import gzip
import os.path
import tempfile

from six import PY2

from ...tests.utils import TemporaryFilename
from .. import utils as io_utils

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def test_gopen():
    # test simple use
    with TemporaryFilename() as tmp:
        with open(tmp, 'w') as f:
            f.write('blah blah blah')
        with io_utils.gopen(tmp) as f2:
            assert f2.read() == 'blah blah blah'

    # test gzip file (with and without extension)
    for suffix in ('.txt.gz', ''):
        with TemporaryFilename(suffix=suffix) as tmp:
            text = b'blah blah blah'
            with gzip.open(tmp, 'wb') as fobj:
                fobj.write(text)
            with io_utils.gopen(tmp, mode='rb') as fobj2:
                assert isinstance(fobj2, gzip.GzipFile)
                assert fobj2.read() == text


def test_identify_factory():
    id_func = io_utils.identify_factory('.blah', '.blah2')
    assert id_func(None, None, None) is False
    assert id_func(None, 'test.txt', None) is False
    assert id_func(None, 'test.blah', None) is True
    assert id_func(None, 'test.blah2', None) is True
    assert id_func(None, 'test.blah2x', None) is False



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for plotter module
"""

import warnings
from io import BytesIO

import pytest

import numpy
from numpy import testing as nptest

from scipy import signal

from matplotlib import (use, rc_context, __version__ as mpl_version)
use('agg')  # nopep8
from matplotlib import (pyplot, rcParams)
from matplotlib.legend import Legend
from matplotlib.lines import Line2D
from matplotlib.colors import (LogNorm, ColorConverter)
from matplotlib.collections import (PathCollection, PatchCollection,
                                    PolyCollection)

from astropy import units

from gwpy.segments import (DataQualityFlag,
                           Segment, SegmentList, SegmentListDict)
from gwpy.frequencyseries import FrequencySeries
from gwpy.timeseries import TimeSeries
from gwpy.table import EventTable
from gwpy.plotter import (figure, Plot, Axes,
                          TimeSeriesPlot, TimeSeriesAxes,
                          FrequencySeriesPlot, FrequencySeriesAxes,
                          EventTablePlot, EventTableAxes,
                          HistogramPlot, HistogramAxes,
                          SegmentPlot, SegmentAxes,
                          SpectrogramPlot, BodePlot)
from gwpy.plotter.rc import (SUBPLOT_WIDTH, SUBPLOT_HEIGHT)
from gwpy.plotter.gps import (GPSTransform, InvertedGPSTransform)
from gwpy.plotter.html import (map_data, map_artist)
from gwpy.plotter.log import CombinedLogFormatterMathtext
from gwpy.plotter.text import (to_string, unit_as_label)
from gwpy.plotter.tex import (float_to_latex, label_to_latex,
                              unit_to_latex, HAS_TEX)
from gwpy.plotter.table import get_column_string

from . import utils

# ignore matplotlib complaining about GUIs
warnings.filterwarnings(
    'ignore', category=UserWarning, message=".*non-GUI backend.*")

# design ZPK for BodePlot test
ZPK = [100], [1], 1e-2
FREQUENCIES, MAGNITUDE, PHASE = signal.bode(ZPK, n=100)

# extract color cycle
COLOR_CONVERTER = ColorConverter()
try:
    COLOR_CYCLE = rcParams['axes.prop_cycle'].by_key()['color']
except KeyError:  # mpl < 1.5
    COLOR0 = COLOR_CONVERTER.to_rgba('b')
else:
    if mpl_version >= '2.0':
        COLOR0 = COLOR_CONVERTER.to_rgba(COLOR_CYCLE[0])
    else:
        COLOR0 = COLOR_CONVERTER.to_rgba('b')

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


@pytest.fixture(scope='function', params=[
    pytest.param(False, id='usetex'),
    pytest.param(True, id='no-tex', marks=pytest.mark.skipif(
        not HAS_TEX, reason='no latex')),
])
def usetex(request):
    """Fixture to test plotting function with and without `usetex`

    Returns
    -------
    usetex : `bool`
        the value of the `text.usetex` rcParams settings
    """
    use_ = request.param
    with rc_context(rc={'text.usetex': use_}):
        yield use_


class PlottingTestBase(object):
    FIGURE_CLASS = Plot
    AXES_CLASS = Axes

    def new(self, **figkwargs):
        """Create a new `Figure` with some `Axes`

        Returns (fig, ax)
        """
        fig = self.FIGURE_CLASS(**figkwargs)
        return fig, fig.gca()

    @property
    def use_tex(self):
        return rcParams['text.usetex']

    def save(self, fig, suffix='.png'):
        fig.save(BytesIO(), format=suffix.lstrip('.'))
        return fig

    def save_and_close(self, fig, suffix='.png'):
        self.save(fig, suffix=suffix)
        fig.close()
        return fig


class TestPlot(PlottingTestBase):
    def test_init(self):
        # test object creation
        fig, ax = self.new()
        assert isinstance(fig, self.FIGURE_CLASS)

        # test added properties
        assert isinstance(fig.colorbars, list)
        assert len(fig.colorbars) == 0
        assert isinstance(fig._coloraxes, list)
        assert len(fig._coloraxes) == 0

        # test gca
        assert isinstance(ax, self.AXES_CLASS)
        self.save_and_close(fig)

    # -- plot rendering -------------------------

    def test_figure(self):
        fig = figure()
        assert isinstance(fig, Plot)
        fig.close()
        fig = figure(FigureClass=self.FIGURE_CLASS)
        assert isinstance(fig, self.FIGURE_CLASS)
        fig.close()

    def test_close(self):
        fig = self.FIGURE_CLASS()
        assert fig.canvas.manager.num in pyplot.get_fignums()
        fig.close()
        assert fig.canvas.manager.num not in pyplot.get_fignums()

    def test_show(self):
        # no idea how to assert that this worked
        fig = self.FIGURE_CLASS()
        fig.show(block=True, warn=False)
        fig.show(block=False, warn=False)

    def test_refresh(self):
        # no idea how to assert that this worked
        fig, ax = self.new()
        fig.refresh()

    def test_auto_refresh(self):
        fig = self.FIGURE_CLASS()
        assert fig.get_auto_refresh() is False

        fig.set_auto_refresh(True)
        assert fig.get_auto_refresh() is True

        fig = self.FIGURE_CLASS(auto_refresh=True)
        assert fig.get_auto_refresh()
        self.save_and_close(fig)

    def test_subplotpars(self):
        # check that dynamic subplotpars gets applied
        fig, ax = self.new(figsize=(12, 4))
        target = SUBPLOT_WIDTH[12] + SUBPLOT_HEIGHT[4]
        sbp = fig.subplotpars
        assert target == (sbp.left, sbp.right, sbp.bottom, sbp.top)

        # check that dynamic subplotpars doesn't get applied if the user
        # overrides any of the settings
        with rc_context(rc={'figure.subplot.left': target[0]*.1}):
            fig, ax = self.new(figsize=(12, 4))
            sbp = fig.subplotpars
            assert sbp.left, target[0]*.1

    # -- plot manipulation ----------------------

    @pytest.mark.parametrize('name, args', [
        ('xlim', (4, 5)),
        ('xlabel', 'test'),
        ('xscale', 'log'),
        ('ylim', (4, 5)),
        ('ylabel', 'test'),
        ('yscale', 'log'),
        ('title', 'test'),
    ])
    def test_axes_methods(self, name, args):
        fig, ax = self.new()

        fig_get = getattr(fig, 'get_%s' % name)
        fig_set = getattr(fig, 'set_%s' % name)
        ax_get = getattr(ax, 'get_%s' % name)

        # note: @axes_methods is deprecated
        with pytest.warns(DeprecationWarning):
            assert fig_get() == ax_get()
        with pytest.warns(DeprecationWarning):
            fig_set(args)
        with pytest.warns(DeprecationWarning):
            assert fig_get() == args
        with pytest.warns(DeprecationWarning):
            assert fig_get() == ax_get()

    @pytest.mark.parametrize('axis', ('x', 'y'))
    def test_log(self, axis):
        fig, ax = self.new()

        # fig.set_xlim(0.1, 10)
        with pytest.warns(DeprecationWarning):
            getattr(fig, 'set_%slim' % axis)(0.1, 10)

        # fig.logx = True
        with pytest.warns(DeprecationWarning):
            setattr(fig, 'log%s' % axis, True)

        # assert ax.get_xscale() == 'log'
        assert getattr(ax, 'get_%sscale' % axis)() == 'log'

        # assert fig.logx is True
        with pytest.warns(DeprecationWarning):
            assert getattr(fig, 'log%s' % axis) is True

        # fig.logx = False
        with pytest.warns(DeprecationWarning):
            setattr(fig, 'log%s' % axis, False)
        assert getattr(ax, 'get_%sscale' % axis)() == 'linear'
        with pytest.warns(DeprecationWarning):
            assert getattr(fig, 'log%s' % axis) is False

        self.save_and_close(fig)

    # -- artist creation ------------------------

    def test_add_legend(self):
        fig, ax = self.new()
        ax.plot([1, 2, 3, 4], label='Plot')
        assert isinstance(fig.add_legend(), Legend)
        self.save_and_close(fig)

    def test_add_line(self):
        fig, ax = self.new()
        fig.add_line([1, 2, 3, 4], [1, 2, 3, 4])
        assert len(ax.lines) == 1
        fig.close()

    def test_add_scatter(self):
        fig, ax = self.new()
        fig.add_scatter([1, 2, 3, 4], [1, 2, 3, 4])
        assert len(ax.collections) == 1
        fig.close()

    def test_add_image(self):
        fig, ax = self.new()
        data = numpy.arange(120).reshape((10, 12))
        image = fig.add_image(data, cmap='Blues')
        assert ax.images == [image]
        assert image.get_cmap().name == 'Blues'

        fig = self.FIGURE_CLASS()
        fig.add_image(data)

    def test_add_arrays(self):
        ts = TimeSeries([1, 2, 3, 4])
        fs = FrequencySeries([1, 2, 3, 4])
        fig = self.FIGURE_CLASS()
        assert len(fig.axes) == 0

        fig.add_timeseries(ts)
        assert len(fig.axes) == 1
        assert isinstance(fig.axes[0], TimeSeriesAxes)

        fig.add_frequencyseries(fs)
        assert len(fig.axes) == 2
        assert isinstance(fig.axes[1], FrequencySeriesAxes)
        fig.close()

    def test_add_colorbar(self):
        fig = self.FIGURE_CLASS()

        # test that adding a colorbar to an empty plot fails
        with pytest.raises(ValueError) as exc:
            fig.add_colorbar()
        assert str(exc.value) == ('Cannot determine mappable layer for '
                                  'this colorbar')
        with pytest.raises(ValueError) as exc:
            fig.add_colorbar(visible=False)
        assert str(exc.value) == ('Cannot determine an anchor Axes for '
                                  'this colorbar')

        # add axes and an image
        ax = fig.gca()
        width = ax.get_position().width
        mappable = ax.imshow(numpy.arange(120).reshape((10, 12)))
        assert not isinstance(mappable.norm, LogNorm)

        # add colorbar and check everything went through
        cbar = fig.add_colorbar(log=True, label='Test label', cmap='YlOrRd')
        assert len(fig.axes) == 2
        assert cbar in fig.colorbars
        assert cbar.ax in fig._coloraxes
        assert cbar.mappable is mappable
        assert cbar.get_clim() == mappable.get_clim() == (1, 119)
        assert isinstance(mappable.norm, LogNorm)
        assert isinstance(cbar.formatter, CombinedLogFormatterMathtext)
        assert cbar.get_cmap().name == 'YlOrRd'
        assert cbar.ax.get_ylabel() == 'Test label'
        self.save_and_close(fig)
        assert ax.get_position().width < width

        # try a non-visible colorbar
        fig = self.FIGURE_CLASS()
        ax = fig.gca()
        assert len(fig.axes) == 1
        fig.add_colorbar(ax=ax, visible=False)
        assert len(fig.axes) == 1
        fig.close()

        # check errors
        mappable = ax.imshow(numpy.arange(120).reshape((10, 12)))
        with pytest.raises(ValueError):
            fig.add_colorbar(mappable=mappable, location='bottom')


class TestAxes(PlottingTestBase):

    # -- test properties ------------------------
    # all of these properties are DEPRECATED

    @pytest.mark.parametrize('axis', ('x', 'y'))
    def test_label(self, axis):
        fig, ax = self.new()

        axis_obj = getattr(ax, '%saxis' % axis)
        label = '%slabel' % axis
        get_label = getattr(ax, 'get_%slabel' % axis)

        # assert ax.xlabel is ax.xaxis.label
        with pytest.warns(DeprecationWarning):
            assert getattr(ax, label) is axis_obj.label

        # ax.xlabel = 'Test label'
        with pytest.warns(DeprecationWarning):
            setattr(ax, label, 'Test label')
        # assert ax.get_xlabel() == 'Test label'
        assert get_label() == 'Test label'

        # check Text object gets preserved
        t = ax.text(0, 0, 'Test text')
        with pytest.warns(DeprecationWarning):
            setattr(ax, label, t)
        assert axis_obj.label is t

        # check deleter works
        with pytest.warns(DeprecationWarning):
            delattr(ax, label)
        assert get_label() == ''

    @pytest.mark.parametrize('axis', ('x', 'y'))
    def test_lim(self, axis):
        fig, ax = self.new()
        ax.plot([1, 2, 3, 4, 5], [1, 2, 3, 4, 5])
        lim = '%slim' % axis
        get_lim = getattr(ax, 'get_%slim' % axis)

        # check getter/setter
        with pytest.warns(DeprecationWarning):
            setattr(ax, lim, (24, 36))
        assert get_lim() == (24, 36)
        with pytest.warns(DeprecationWarning):
            assert getattr(ax, lim) == get_lim()

        # check deleter
        with pytest.warns(DeprecationWarning):
            delattr(ax, lim)

    @pytest.mark.parametrize('axis', ('x', 'y'))
    def test_log(self, axis):
        fig, ax = self.new()
        log = 'log%s' % axis
        get_scale = getattr(ax, 'get_%sscale' % axis)
        set_scale = getattr(ax, 'set_%sscale' % axis)

        # check default is not log
        with pytest.warns(DeprecationWarning):
            assert getattr(ax, log) is False

        # set log and assert that the scale gets set properly
        with pytest.warns(DeprecationWarning):
            setattr(ax, log, True)
        with pytest.warns(DeprecationWarning):
            assert getattr(ax, log) is True
        assert get_scale() == 'log'

        # set not log and check
        with pytest.warns(DeprecationWarning):
            setattr(ax, log, False)
        with pytest.warns(DeprecationWarning):
            assert getattr(ax, log) is False
        assert get_scale() == 'linear'

    # -- test methods ---------------------------

    def test_resize(self):
        fig, ax = self.new()
        ax.resize((0.25, 0.25, 0.5, 0.5))
        assert ax.get_position().bounds == (.25, .25, .5, .5)

    def test_legend(self):
        fig = self.FIGURE_CLASS()
        ax = fig.add_subplot(111, projection=self.AXES_CLASS.name)
        ax.plot([1, 2, 3, 4], label='Plot')

        leg = ax.legend()
        legframe = leg.get_frame()
        assert legframe.get_alpha() == 0.8
        assert legframe.get_linewidth() == rcParams['axes.linewidth']

        for l in leg.get_lines():
            assert l.get_linewidth() == 8
        self.save_and_close(fig)

    def test_html_map(self):
        # this method just runs the html_map method but puts little effort
        # into validating the result, that is left for the TestHtml
        # suite

        fig, ax = self.new()

        with pytest.raises(ValueError) as exc:
            ax.html_map('test.png')
        assert str(exc.value) == 'Cannot determine artist to map, 0 found.'

        line = ax.plot([1, 2, 3, 4, 5])[0]

        # auto-detect artist
        hmap = ax.html_map('test.png')
        assert hmap.startswith('<!doctype html>')
        assert hmap.count('<area') == 5

        # manually pass artist
        hmap2 = ax.html_map('test.png', data=line)
        assert hmap2 == hmap

        # auto-detect with multiple artists
        ax.plot([1, 2, 3, 4, 5])
        with pytest.raises(ValueError):
            ax.html_map('test.png')

        # check data=<array>
        data = list(zip(*line.get_data()))
        hmap2 = ax.html_map('test.png', data=data)
        assert hmap2 == hmap

    @pytest.mark.parametrize('method', [
        'plot',
        'loglog',
        'semilogx',
        'semilogy',
    ])
    def test_matplotlib_compat(self, method):
        fig, ax = self.new()
        plot = getattr(ax, method)
        plot([1, 2, 3, 4, 5])
        plot([1, 2, 3, 4, 5], color='green', linestyle='--')
        self.save_and_close(fig)


# -- TimeSeries plotters ------------------------------------------------------

class TimeSeriesMixin(object):
    FIGURE_CLASS = TimeSeriesPlot
    AXES_CLASS = TimeSeriesAxes

    @classmethod
    def setup_class(cls):
        numpy.random.seed(0)
        cls.ts = TimeSeries(numpy.random.rand(10000), sample_rate=128)
        cls.sg = cls.ts.spectrogram2(0.5, 0.49)
        cls.mmm = [cls.ts, cls.ts * 0.9, cls.ts*1.1]


class TestTimeSeriesPlot(TimeSeriesMixin, TestPlot):
    def test_init(self):
        # test empty
        fig, ax = self.new()
        assert isinstance(ax, self.AXES_CLASS)

        # test passing arguments
        fig = self.FIGURE_CLASS(figsize=[9, 6])
        assert fig.get_figwidth() == 9
        fig = self.FIGURE_CLASS(self.ts)
        ax = fig.gca()
        assert len(ax.lines) == 1
        assert ax.get_xlim() == self.ts.span

        # test passing multiple timeseries
        fig = self.FIGURE_CLASS(self.ts, self.ts)
        assert len(fig.gca().lines) == 2
        fig = self.FIGURE_CLASS(self.ts, self.ts, sep=True, sharex=True)
        assert len(fig.axes) == 2
        for ax in fig.axes:
            assert len(ax.lines) == 1
        assert fig.axes[1]._sharex is fig.axes[0]

        # test kwarg parsing
        fig = self.FIGURE_CLASS(self.ts, figsize=[12, 6], rasterized=True)

    def test_add_colorbar(self):
        def make_fig():
            fig = self.FIGURE_CLASS(self.sg ** (1/2.))
            return fig, fig.gca()

        # test basic
        fig, ax = make_fig()
        cb = fig.add_colorbar()
        assert len(fig.colorbars) == 1
        assert cb is fig.colorbars[0]
        self.save_and_close(fig)
        # test kwargs
        fig, ax = make_fig()
        fig.add_colorbar(ax=ax, norm='log', clim=[1e-22, 1e-18],
                         label='Test colorbar')
        self.save_and_close(fig)

    def test_add_state_segments(self):
        fig, ax = self.new()

        # mock up some segments and add them as 'state' segments
        segs = SegmentList([Segment(1, 2), Segment(4, 5)])
        segax = fig.add_state_segments(segs)

        # check that the new axes aligns with the parent
        utils.assert_array_equal(segax.get_position().intervalx,
                                 ax.get_position().intervalx)
        coll = segax.collections[0]
        for seg, path in zip(segs, coll.get_paths()):
            utils.assert_array_equal(
                path.vertices, [(seg[0], -.4), (seg[1], -.4), (seg[1], .4),
                                (seg[0], .4), (seg[0], -.4)])

        with pytest.raises(ValueError):
            fig.add_state_segments(segs, location='left')

        # test that this doesn't work with non-timeseries axes
        fig = self.FIGURE_CLASS()
        ax = fig.gca(projection='rectilinear')
        with pytest.raises(ValueError) as exc:
            fig.add_state_segments(segs)
        assert str(exc.value) == ("No 'timeseries' Axes found, cannot anchor "
                                  "new segment Axes.")


class TestTimeSeriesAxes(TimeSeriesMixin, TestAxes):
    def test_init(self):
        fig, ax = self.new()
        assert isinstance(ax, self.AXES_CLASS)
        assert ax.get_xscale() == 'auto-gps'
        assert ax.get_xlabel() == '_auto'
        self.save_and_close(fig)

    def test_plot_timeseries(self):
        fig, ax = self.new()
        # check method works
        ax.plot_timeseries(self.ts)
        # check data are correct
        line = ax.get_lines()[0]
        nptest.assert_array_equal(line.get_xdata(), self.ts.times.value)
        nptest.assert_array_equal(line.get_ydata(), self.ts.value)
        # check GPS axis is set ok
        assert ax.get_xlim() == tuple(self.ts.span)
        self.save_and_close(fig)

    def test_plot_mmm(self):
        fig, ax = self.new()
        # test default
        a, b, c, d, e = ax.plot_mmm(*self.mmm)
        for line in (a, b, d):
            assert isinstance(line, Line2D)
        for coll in (c, e):
            assert isinstance(coll, PolyCollection)
        assert len(ax.lines) == 3
        assert len(ax.collections) == 2
        self.save_and_close(fig)

        # test with labels
        fig, ax = self.new()
        minname = self.mmm[1].name
        maxname = self.mmm[2].name
        self.mmm[1].name = 'min'
        self.mmm[2].name = 'max'
        try:
            ax.plot_mmm(*self.mmm, label='test')
            leg = ax.legend()
            assert len(leg.get_lines()) == 1
        finally:
            self.mmm[1].name = minname
            self.mmm[2].name = maxname
        self.save_and_close(fig)

        # test min only
        fig, ax = self.new()
        a, b, c, d, e = ax.plot_mmm(self.mmm[0], min_=self.mmm[1])
        assert d is None
        assert e is None
        assert len(ax.lines) == 2
        assert len(ax.collections) == 1
        self.save_and_close(fig)

        # test max only
        fig, ax = self.new()
        a, b, c, d, e = ax.plot_mmm(self.mmm[0], max_=self.mmm[2])
        assert b is None
        assert c is None
        assert len(ax.lines) == 2
        assert len(ax.collections) == 1

    def test_plot_timeseries_mmm(self):
        fig, ax = self.new()
        with pytest.warns(DeprecationWarning):
            ax.plot_timeseries_mmm(*self.mmm)
        self.save_and_close(fig)

    def test_plot_spectrogram(self):
        fig, ax = self.new()
        # check method
        ax.plot_spectrogram(self.sg, imshow=False)
        coll = ax.collections[0]
        nptest.assert_array_equal(coll.get_array(), self.sg.value.T.flatten())
        # check GPS axis is set ok
        assert ax.get_xlim() == tuple(self.sg.xspan)
        # check frequency axis
        if self.use_tex:
            assert ax.get_ylabel() == r'Frequency [$\mathrm{Hz}$]'
        else:
            assert ax.get_ylabel() == r'Frequency [Hz]'
        # check kwarg parsing
        c = ax.plot_spectrogram(self.sg, norm='log')
        assert isinstance(c.norm, LogNorm)
        self.save_and_close(fig)

    def test_plot(self):
        fig, ax = self.new()
        ax.plot(self.ts)
        self.save_and_close(fig)


# -- FrequencySeries plotters -------------------------------------------------

class FrequencySeriesMixin(object):
    FIGURE_CLASS = FrequencySeriesPlot
    AXES_CLASS = FrequencySeriesAxes

    @classmethod
    def setup_class(cls):
        numpy.random.seed(0)
        cls.ts = TimeSeries(numpy.random.rand(10000), sample_rate=128)
        cls.asd = cls.ts.asd(1)
        cls.mmm = [cls.asd, cls.asd*0.9, cls.asd*1.1]


class TestFrequencySeriesPlot(FrequencySeriesMixin, TestPlot):
    def test_init(self):
        super(TestFrequencySeriesPlot, self).test_init()
        # test convenience plotting
        fig = self.FIGURE_CLASS(self.asd)
        assert len(fig.axes) == 1
        ax = fig.gca()
        assert isinstance(ax, self.AXES_CLASS)
        assert len(ax.lines) == 1


class TestFrequencySeriesAxes(FrequencySeriesMixin, TestAxes):
    def test_plot_frequencyseries(self):
        fig, ax = self.new()
        line = ax.plot_frequencyseries(self.asd)[0]
        nptest.assert_array_equal(line.get_xdata(), self.asd.frequencies.value)
        nptest.assert_array_equal(line.get_ydata(), self.asd.value)
        self.save_and_close(fig)

    def test_plot_mmm(self):
        fig, ax = self.new()
        # test defaults
        a, b, c, d, e = ax.plot_mmm(*self.mmm)
        for line in (a, b, d):
            assert isinstance(line, Line2D)
        for coll in (c, e):
            assert isinstance(coll, PolyCollection)
        assert len(ax.lines) == 3
        assert len(ax.collections) == 2
        self.save_and_close(fig)

        # test with labels
        fig, ax = self.new()
        minname = self.mmm[1].name
        maxname = self.mmm[2].name
        self.mmm[1].name = 'min'
        self.mmm[2].name = 'max'
        try:
            ax.plot_mmm(*self.mmm, label='test')
            leg = ax.legend()
            assert len(leg.get_lines()) == 1
        finally:
            self.mmm[1].name = minname
            self.mmm[2].name = maxname
        self.save_and_close(fig)

        # test min only
        fig, ax = self.new()
        a, b, c, d, e = ax.plot_mmm(self.mmm[0], min_=self.mmm[1])
        assert d is None
        assert e is None
        assert len(ax.lines) == 2
        assert len(ax.collections) == 1
        self.save_and_close(fig)

        # test max only
        fig, ax = self.new()
        a, b, c, d, e = ax.plot_mmm(self.mmm[0], max_=self.mmm[2])
        assert b is None
        assert c is None
        assert len(ax.lines) == 2
        assert len(ax.collections) == 1
        self.save_and_close(fig)

    def test_plot_frequencyseries_mmm(self):
        fig, ax = self.new()
        with pytest.warns(DeprecationWarning):
            ax.plot_frequencyseries_mmm(*self.mmm)
        self.save_and_close(fig)


# -- Table plotters -----------------------------------------------------------

class EventTableMixin(object):
    FIGURE_CLASS = EventTablePlot
    AXES_CLASS = EventTableAxes

    @classmethod
    def create(cls, n, names, dtypes=None):
        data = []
        for i in range(len(names)):
            numpy.random.seed(i)
            if dtypes:
                dtype = dtypes[i]
            else:
                dtype = None
            data.append((numpy.random.rand(n) * 1000).astype(dtype))
        return EventTable(data, names=names)

    @classmethod
    @pytest.fixture()
    def table(cls):
        return cls.create(100, ['time', 'snr', 'frequency',
                                'duration', 'bandwidth'])


class TestEventTablePlot(EventTableMixin, TestPlot):
    def test_init_with_table(self, table):
        self.FIGURE_CLASS(table, 'time', 'frequency').close()
        with pytest.raises(ValueError):
            self.FIGURE_CLASS(table)
        self.FIGURE_CLASS(
            table, 'time', 'frequency', 'snr').close()
        self.FIGURE_CLASS(
            table, 'time', 'frequency', 'snr').close()


class TestEventTableAxes(EventTableMixin, TestAxes):
    def test_plot_table(self, table):
        fig, ax = self.new()
        snrs = table.get_column('snr')
        snrs.sort()
        # test with color
        c = ax.plot_table(table, 'time', 'frequency', 'snr')
        shape = c.get_offsets().shape
        assert isinstance(c, PathCollection)
        assert shape[0] == len(table)
        nptest.assert_array_equal(c.get_array(), snrs)
        # test with size_by
        with pytest.warns(DeprecationWarning):
            c = ax.plot_table(table, 'time', 'frequency', size_by='snr')
        # test with color and size_by
        with pytest.warns(DeprecationWarning):
            c = ax.plot_table(table, 'time', 'frequency', 'snr', size_by='snr')
        nptest.assert_array_equal(c.get_array(), snrs)

    def test_plot_tiles(self, table):
        fig, ax = self.new()
        snrs = table.get_column('snr')
        snrs.sort()
        # test with color
        c = ax.plot_tiles(table, 'time', 'frequency', 'duration',
                          'bandwidth', 'snr')
        assert isinstance(c, PolyCollection)
        # test other anchors
        c = ax.plot_tiles(table, 'time', 'frequency', 'duration',
                          'bandwidth', 'snr', anchor='ll')
        c = ax.plot_tiles(table, 'time', 'frequency', 'duration',
                          'bandwidth', 'snr', anchor='lr')
        c = ax.plot_tiles(table, 'time', 'frequency', 'duration',
                          'bandwidth', 'snr', anchor='ul')
        c = ax.plot_tiles(table, 'time', 'frequency', 'duration',
                          'bandwidth', 'snr', anchor='ur')
        with pytest.raises(ValueError):
            ax.plot_tiles(table, 'time', 'frequency', 'duration',
                          'bandwidth', 'snr', anchor='other')

    def test_get_column_string(self):
        with rc_context(rc={'text.usetex': True}):
            assert get_column_string('snr') == 'SNR'
            assert get_column_string('reduced_chisq') == r'Reduced $\chi^2$'
            assert get_column_string('flow') == r'f$_{\mbox{\small low}}$'
            assert get_column_string('end_time_ns') == r'End Time $(ns)$'

    def test_add_loudest(self, usetex, table):
        table.add_column(table.Column(data=['test'] * len(table), name='test'))
        loudest = table[table['snr'].argmax()]
        t, f, s = loudest['time'], loudest['frequency'], loudest['snr']

        # make plot
        fig, ax = self.new()
        ax.scatter(table['time'], table['frequency'])
        tpos = ax.title.get_position()

        # call function
        coll, text = ax.add_loudest(
            table, 'snr',  # table, rank
            'time', 'frequency',  # x, y
            'test',  # extra columns to print
            'time',  # duplicate (shouldn't get printed)
        )

        # check marker was placed at the right point
        utils.assert_array_equal(coll.get_offsets(), [(t, f)])

        # check text
        result = ('Loudest event: Time = {0}, Frequency = {1}, SNR = {2}, '
                  'Test = test'.format(
                      *('{0:.2f}'.format(x) for x in (t, f, s))))

        assert text.get_text() == result
        assert text.get_position() == (.5, 1.)

        # assert title got moved
        assert ax.title.get_position() == (tpos[0], tpos[1] + .05)

        # -- with more kwargs

        _, t = ax.add_loudest(table, 'snr', 'time', 'frequency',
                              position=(0., 0.), ha='left', va='top')
        assert t.get_position() == (0., 0.)

        # assert title doesn't get moved again if we specify position
        assert ax.title.get_position() == (tpos[0], tpos[1] + .05)

        # assert kw handling
        assert t.get_horizontalalignment() == 'left'
        assert t.get_verticalalignment() == 'top'

        self.save_and_close(fig)


# -- Segment plotter ----------------------------------------------------------

class SegmentMixin(object):
    FIGURE_CLASS = SegmentPlot
    AXES_CLASS = SegmentAxes

    @staticmethod
    @pytest.fixture()
    def segments():
        return SegmentList([Segment(0, 3), Segment(6, 7)])

    @staticmethod
    @pytest.fixture()
    def flag():
        known = SegmentList([Segment(0, 3), Segment(6, 7)])
        active = SegmentList([Segment(1, 2), Segment(3, 4), Segment(5, 7)])
        return DataQualityFlag(name='Test segments', known=known,
                               active=active)


class TestSegmentPlot(SegmentMixin, TestPlot):
    def test_add_bitmask(self, segments):
        fig, ax = self.new()
        ax.plot(segments)
        maxes = fig.add_bitmask(0b0110)
        assert maxes.name == ax.name  # test same type
        maxes = fig.add_bitmask('0b0110', topdown=True)


class TestSegmentAxes(SegmentMixin, TestAxes):

    def test_plot_dqflag(self, flag):
        fig, ax = self.new()
        c = ax.plot_dqflag(flag)
        assert c.get_label() == flag.texname
        assert len(ax.collections) == 2
        assert ax.collections[1] is c

    def test_build_segment(self):
        patch = self.AXES_CLASS.build_segment((1.1, 2.4), 10)
        assert patch.get_xy(), (1.1, 9.6)
        assert numpy.isclose(patch.get_height(), 0.8)
        assert numpy.isclose(patch.get_width(), 1.3)
        assert patch.get_facecolor(), COLOR0
        # check kwarg passing
        patch = self.AXES_CLASS.build_segment((1.1, 2.4), 10, facecolor='red')
        assert patch.get_facecolor() == COLOR_CONVERTER.to_rgba('red')
        # check valign
        patch = self.AXES_CLASS.build_segment((1.1, 2.4), 10, valign='top')
        assert patch.get_xy() == (1.1, 9.2)
        patch = self.AXES_CLASS.build_segment((1.1, 2.4), 10, valign='bottom')
        assert patch.get_xy() == (1.1, 10.0)

    def test_plot_segmentlist(self, segments):
        fig, ax = self.new()
        c = ax.plot_segmentlist(segments)
        assert isinstance(c, PatchCollection)
        assert numpy.isclose(ax.dataLim.x0, 0.)
        assert numpy.isclose(ax.dataLim.x1, 7.)
        assert len(c.get_paths()) == len(segments)
        # test y
        p = ax.plot_segmentlist(segments).get_paths()[0].get_extents()
        assert p.y0 + p.height/2. == 1.
        p = ax.plot_segmentlist(segments, y=8).get_paths()[0].get_extents()
        assert p.y0 + p.height/2. == 8.
        # test kwargs
        c = ax.plot_segmentlist(segments, label='My segments',
                                rasterized=True)
        assert c.get_label() == 'My segments'
        assert c.get_rasterized() is True
        # test collection=False
        c = ax.plot_segmentlist(segments, collection=False, label='test')
        assert isinstance(c, list)
        assert not isinstance(c, PatchCollection)
        assert c[0].get_label() == 'test'
        assert c[1].get_label() == ''
        assert len(ax.patches) == len(segments)
        # test empty
        c = ax.plot_segmentlist(type(segments)())
        self.save_and_close(fig)

    def test_plot_segmentlistdict(self, segments):
        sld = SegmentListDict()
        sld['TEST'] = segments
        fig, ax = self.new()
        ax.plot(sld)
        self.save_and_close(fig)

    def test_plot(self, segments, flag):
        fig, ax = self.new()
        ax.plot(segments)
        ax.plot(flag)
        ax.plot(flag, segments)
        self.save_and_close(fig)

    def test_insetlabels(self, segments):
        fig, ax = self.new()
        ax.plot(segments)
        ax.set_insetlabels(True)
        self.save_and_close(fig)


# -- Histogram plotter --------------------------------------------------------

class HistogramMixin(object):
    FIGURE_CLASS = HistogramPlot
    AXES_CLASS = HistogramAxes

    @classmethod
    def setup_class(cls):
        numpy.random.seed(0)
        cls.ts = TimeSeries(numpy.random.rand(10000), sample_rate=128)


class TestHistogramPlot(HistogramMixin, TestPlot):
    def test_init(self):
        super(TestHistogramPlot, self).test_init()

        # test with data
        bins = [1, 2, 5]
        plot = self.FIGURE_CLASS([1, 2, 3, 4, 5], bins=bins)
        assert len(plot.axes) == 1
        ax = plot.gca()
        assert len(ax.containers) == 1
        assert len(ax.containers[0].patches) == len(bins) - 1


class TestHistogramAxes(HistogramMixin, TestAxes):
    def test_hist_series(self):
        fig, ax = self.new()
        ax.hist_series(self.ts)

    def test_hist(self):
        fig, ax = self.new()
        ax.hist(self.ts)

    @utils.skip_missing_dependency('glue')
    def test_common_limits(self):
        fig, ax = self.new()
        a = ax.common_limits(self.ts.value)
        assert numpy.allclose(a, (7.2449638492178003e-05, 0.9999779517807228))
        b = ax.common_limits([self.ts.value])
        assert a == b

    def test_bin_boundaries(self):
        ax = self.AXES_CLASS
        nptest.assert_array_equal(ax.bin_boundaries(0, 10, 4),
                                  [0., 2.5, 5., 7.5, 10.])
        nptest.assert_array_equal(ax.bin_boundaries(1, 100, 2, log=True),
                                  [1, 10., 100.])
        with pytest.raises(ValueError):
            ax.bin_boundaries(0, 100, 2, log=True)

    def test_histogram_weights(self):
        fig, ax = self.new()
        ax.hist(numpy.random.random(1000), weights=10.)
        self.save_and_close(fig)


# -- Filter plotter -----------------------------------------------------------

class TestBodePlot(PlottingTestBase):
    FIGURE_CLASS = BodePlot

    def test_init(self):
        fig = self.FIGURE_CLASS()
        assert len(fig.axes) == 2
        maxes, paxes = fig.axes
        # test magnigtude axes
        assert isinstance(maxes, FrequencySeriesAxes)
        assert maxes.get_xscale() == 'log'
        assert maxes.get_xlabel() == ''
        assert maxes.get_yscale() == 'linear'
        assert maxes.get_ylabel() == 'Magnitude [dB]'
        # test phase axes
        assert isinstance(paxes, FrequencySeriesAxes)
        assert paxes.get_xscale() == 'log'
        assert paxes.get_xlabel() == 'Frequency [Hz]'
        assert paxes.get_yscale() == 'linear'
        assert paxes.get_ylabel() == 'Phase [deg]'

    def test_add_filter(self):
        # test method 1
        fig = self.FIGURE_CLASS()
        lm, lp = fig.add_filter(ZPK, analog=True)
        assert lm is fig.maxes.get_lines()[-1]
        assert lp is fig.paxes.get_lines()[-1]
        nptest.assert_array_equal(lm.get_xdata(), FREQUENCIES)
        nptest.assert_array_equal(lm.get_ydata(), MAGNITUDE)
        nptest.assert_array_equal(lp.get_xdata(), FREQUENCIES)
        nptest.assert_array_almost_equal(lp.get_ydata(), PHASE)
        self.save_and_close(fig)
        # test method 2
        fig = self.FIGURE_CLASS(ZPK, analog=True)
        lm = fig.maxes.get_lines()[0]
        lp = fig.paxes.get_lines()[0]
        nptest.assert_array_equal(lm.get_xdata(), FREQUENCIES)
        nptest.assert_array_equal(lm.get_ydata(), MAGNITUDE)
        nptest.assert_array_equal(lp.get_xdata(), FREQUENCIES)
        nptest.assert_array_almost_equal(lp.get_ydata(), PHASE)


# -- gwpy.plotter.gps module tests --------------------------------------------

class TestGpsTransform(object):
    TRANSFORM = GPSTransform
    EPOCH = 100.0
    UNIT = 'minutes'
    SCALE = 60.
    X = 190.0
    A = 90.0
    B = 19/6.
    C = 1.5

    def test_empty(self):
        t = self.TRANSFORM()
        assert t.transform(1.0) == 1.0

    def test_epoch(self):
        transform = self.TRANSFORM(epoch=self.EPOCH)
        assert transform.get_epoch() == self.EPOCH
        assert transform.transform(self.X) == self.A
        assert numpy.isclose(
            transform.inverted().transform(transform.transform(self.X)),
            self.X)

    def test_scale(self):
        transform = self.TRANSFORM(unit=self.UNIT)
        assert transform.get_scale() == self.SCALE
        assert transform.transform(self.X) == self.B
        assert numpy.isclose(
            transform.inverted().transform(transform.transform(self.X)),
            self.X)

    def test_epoch_and_scale(self):
        transform = self.TRANSFORM(epoch=self.EPOCH, unit=self.UNIT)
        assert transform.transform(self.X) == self.C
        assert numpy.isclose(
            transform.inverted().transform(transform.transform(self.X)),
            self.X)


class TestInverseGpsTransform(TestGpsTransform):
    TRANSFORM = InvertedGPSTransform
    A = 290.0
    B = 11400.0
    C = 11500.0


# -- gwpy.plotter.text module tests -------------------------------------------

class TestText(PlottingTestBase):
    def test_to_string(self):
        # test without latex
        with rc_context(rc={'text.usetex': False}):
            assert to_string('test') == 'test'
            assert to_string(4.0) == '4.0'
            assert to_string(8) == '8'
        with rc_context(rc={'text.usetex': True}):
            assert to_string('test') == 'test'
            assert to_string(2000) == r'2\!\!\times\!\!10^{3}'
            assert to_string(8) == '8'

    def test_unit_as_label(self):
        # just test basics, latex formatting is tested elsewhere
        with rc_context(rc={'text.usetex': False}):
            assert unit_as_label(units.Hz) == 'Frequency [Hz]'
            assert unit_as_label(units.Volt) == 'Electrical Potential [V]'


# -- gwpy.plotter.tex module tests --------------------------------------------

class TestTex(PlottingTestBase):
    def test_float_to_latex(self):
        assert float_to_latex(1) == '1'
        assert float_to_latex(100) == '10^{2}'
        assert float_to_latex(-500) == r'-5\!\!\times\!\!10^{2}'
        with pytest.raises(TypeError):
            float_to_latex('1')

    def test_label_to_latex(self):
        assert label_to_latex(None) == ''
        assert label_to_latex('') == ''
        assert label_to_latex('Test') == 'Test'
        assert label_to_latex('Test_with_underscore') == (
            r'Test\_with\_underscore')
        assert label_to_latex(r'Test_with\_escaped\%characters') == (
            r'Test\_with\_escaped\%characters')

    def test_unit_to_latex(self):
        t = unit_to_latex(units.Hertz)
        assert t == r'$\mathrm{Hz}$'
        t = unit_to_latex(units.Volt.decompose())
        assert t == r'$\mathrm{m^{2}\,kg\,A^{-1}\,s^{-3}}$'


# -- gwpy.plotter.html module tests -------------------------------------------

class TestHtml(object):
    def test_map_data(self):
        numpy.random.seed(0)
        data = numpy.vstack((numpy.arange(100), numpy.random.random(100)))
        fig = figure()
        ax = fig.gca()
        html = map_data(data, ax, 'test.png')
        assert html.startswith('<!doctype html>')
        html = map_data(data, ax, 'test.png', standalone=False)
        assert html.startswith('\n<img src="test.png"')

    @utils.skip_missing_dependency('bs4')
    def test_map_artist(self):
        from bs4 import BeautifulSoup

        # create figure and plot a line
        fig = figure()
        ax = fig.gca()
        line = ax.plot([1, 2, 3, 4, 5])[0]
        data = list(zip(*line.get_data()))

        # create HTML map
        html = map_artist(line, 'test.png')

        # validate HTML map
        soup = BeautifulSoup(html, 'html.parser')
        areas = soup.find_all('area')
        assert len(areas) == 5
        assert sorted([eval(a.attrs['alt']) for a in areas]) == data



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.utils.misc`
"""

import sys

import pytest

from .. import misc as utils_misc


def test_gprint(capsys):
    """Test for :func:`gwpy.utils.misc.gprint`
    """
    utils_misc.gprint('test')
    assert capsys.readouterr().out == 'test\n'
    utils_misc.gprint('test', end=' ')
    assert capsys.readouterr().out == 'test '
    utils_misc.gprint('test', end='x', file=sys.stderr)
    cap = capsys.readouterr()
    assert not cap.out
    assert cap.err == 'testx'


def test_null_context():
    """Test for :func:`gwpy.utils.misc.null_context`
    """
    ctx = utils_misc.null_context()
    with ctx:
        print('this should work')


@pytest.mark.parametrize('func, value, out', [
    (str, None, None),
    (str, 1, '1'),
])
def test_if_not_none(func, value, out):
    """Test for :func:`gwpy.utils.misc.if_not_none`
    """
    assert utils_misc.if_not_none(func, value) == out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2018)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Test setup for gwpy
"""

import warnings

import numpy

from matplotlib import use

# ignore errors due from pyplot.show() using Agg
warnings.filterwarnings('ignore', message=".*non-GUI backend.*")

# set random seed to 1 for reproducability
numpy.random.seed(1)

# force Agg for all tests
use('agg', warn=False, force=True)



# -*- coding: utf-8 -*-
# Copyright (C) Scott Coughlin (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extend :mod:`astropy.table` with the `GravitySpyTable`
"""

import os

from ..utils import mp as mp_utils
from .table import EventTable
import numpy as np

__author__ = 'Scott Coughlin <scott.coughlin@ligo.org>'
__all__ = ['GravitySpyTable']


class GravitySpyTable(EventTable):
    """A container for a table of Gravity Spy Events (as well as
    Events from the O1 Glitch Classification Paper whcih includes

    - PCAT
    - PC-LIB
    - WDF
    - WDNN
    - Karoo GP

    This differs from the basic `~astropy.table.Table` in two ways

    - GW-specific file formats are registered to use with
      `GravitySpyTable.fetch`
    - columns of this table are of the `EventColumn` type, which provides
      methods for filtering based on a `~gwpy.segments.SegmentList` (not
      specifically time segments)

    See also
    --------
    astropy.table.Table
        for details on parameters for creating an `GravitySpyTable`
    """

    # -- i/o ------------------------------------

    def download(self, **kwargs):
        """If table contains Gravity Spy triggers `EventTable`

        Parameters
        ----------
        nproc : `int`, optional, default: 1
            number of CPUs to use for parallel file reading

        download_path : `str` optional, default: 'download'
            Specify where the images end up.

        download_durs : `list` optional, default: [0.5, 1.0, 2.0, 4.0]
            Specify exactly which durations you want to download
            default is to download all the avaialble GSpy durations.

        kwargs: Optional TrainingSet and LabelledSamples args
            that will download images in a special way
            ./"Label"/"SampleType"/"image"

        Returns
        -------
        Folder containing omega scans sorted by label
        """
        from six.moves.urllib.request import urlopen
        import os
        # back to pandas
        try:
            imagesDB = self.to_pandas()
        except ImportError as exc:
            exc.args = ('pandas is required to download triggers',)
            raise

        # Remove any broken links
        imagesDB = imagesDB.loc[imagesDB.imgUrl1 != '?']

        TrainingSet = kwargs.pop('TrainingSet', 0)
        LabelledSamples = kwargs.pop('LabelledSamples', 0)
        download_location = kwargs.pop('download_path',
                                       os.path.join('download'))
        duration_values = np.array([0.5, 1.0, 2.0, 4.0])
        download_durs = kwargs.pop('download_durs', duration_values)

        duration_idx = []
        for idur in download_durs:
            duration_idx.append(np.argwhere(duration_values == idur)[0][0])

        duration_values = duration_values[duration_idx]
        duration_values = np.array([duration_values]).astype(str)

        # LabelledSamples are only available when requesting the
        if LabelledSamples:
            if 'SampleType' not in imagesDB.columns:
                raise ValueError('You have requested Labelled Samples '
                                 'for a Table which does not have '
                                 'this column. Did you fetch a '
                                 'trainingset* table?')

        # If someone wants labelled samples they are
        # Definitely asking for the training set but
        # may hve forgotten
        if LabelledSamples and not TrainingSet:
            TrainingSet = 1

        # Let us check what columns are needed
        cols_for_download = ['imgUrl1', 'imgUrl2', 'imgUrl3', 'imgUrl4']
        cols_for_download = [cols_for_download[idx] for idx in duration_idx]
        cols_for_download_ext = ['Label', 'SampleType', 'ifo', 'uniqueID']

        if not TrainingSet:
            imagesDB['Label'] = ''
        if not LabelledSamples:
            imagesDB['SampleType'] = ''

        if not os.path.isdir(download_location):
            os.makedirs(download_location)

        if TrainingSet:
            for iLabel in imagesDB.Label.unique():
                if LabelledSamples:
                    for iType in imagesDB.SampleType.unique():
                        if not os.path.isdir(os.path.join(
                                             download_location,
                                             iLabel, iType)):
                            os.makedirs(os.path.join(download_location,
                                        iLabel, iType))
                else:
                    if not os.path.isdir(os.path.join(download_location,
                                                      iLabel)):
                        os.makedirs(os.path.join(download_location,
                                                 iLabel))

        images_for_download = imagesDB[cols_for_download]
        images = images_for_download.as_matrix().flatten()
        images_for_download_ext = imagesDB[cols_for_download_ext]
        duration = np.atleast_2d(
                                 duration_values.repeat(
                                   len(images_for_download_ext), 0).flatten(
                                     )).T
        images_for_download_ext = images_for_download_ext.as_matrix(
                                       ).repeat(len(cols_for_download), 0)
        images_for_for_download_path = np.array([[download_location]]).repeat(
                                       len(images_for_download_ext), 0)
        images = np.hstack((np.atleast_2d(images).T,
                           images_for_download_ext, duration,
                           images_for_for_download_path))

        def get_image(url):
            name = url[3] + '_' + url[4] + '_spectrogram_' + url[5] + '.png'
            outfile = os.path.join(url[6], url[1], url[2], name)
            with open(outfile, 'wb') as fout:
                fout.write(urlopen(url[0]).read())

        # calculate maximum number of processes
        nproc = min(kwargs.pop('nproc', 1), len(images))

        # define multiprocessing method
        def _download_single_image(url):
            try:
                return url, get_image(url)
            except Exception as exc:  # pylint: disable=broad-except
                if nproc == 1:
                    raise
                else:
                    return url, exc

        # read files
        output = mp_utils.multiprocess_with_queues(
            nproc, _download_single_image, images)

        # raise exceptions (from multiprocessing, single process raises inline)
        for f, x in output:
            if isinstance(x, Exception):
                x.args = ('Failed to read %s: %s' % (f, str(x)),)
                raise x

    @classmethod
    def search(cls, uniqueID, howmany=10):
        """perform restful API version of search available here:
        https://gravityspytools.ciera.northwestern.edu/search/

        Parameters
        ----------
        uniqueID : `str`,
            This is the unique 10 character hash that identifies
            a Gravity Spy Image

        howmany : `int`, optional, default: 10
            number of similar images you would like

        Returns
        -------
        `GravitySpyTable` containing similar events based on
        an evaluation of the Euclidean distance of the input image
        to all other images in some Feature Space
        """
        from astropy.utils.data import get_readable_fileobj
        import json
        from six.moves.urllib.error import HTTPError

        # Need to build the url call for the restful API
        base = 'https://gravityspytools.ciera.northwestern.edu' + \
            '/search/similarity_search_restful_API'

        parts = {
            'howmany': howmany,
            'imageid': uniqueID,
        }

        search = '&'.join('{}={}'.format(key, value) for
                          key, value in parts.items())

        url = '{}/?{}'.format(base, search)

        try:
            with get_readable_fileobj(url) as f:
                return GravitySpyTable(json.load(f))
        except HTTPError as exc:
            if exc.code == 500:
                exc.msg = exc.msg + ', please confirm the uniqueID is valid'
                raise



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for frequencyseries module
"""

import tempfile
from io import BytesIO

import pytest

import numpy

try:
    from numpy import shares_memory
except ImportError:  # old numpy
    from numpy import may_share_memory as shares_memory

from scipy import signal

from matplotlib import rc_context

from astropy import units

from ...tests import utils
from ...timeseries import TimeSeries
from ...types.tests.test_series import TestSeries as _TestSeries
from .. import FrequencySeries

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

LIGO_LW_ARRAY = """<?xml version='1.0' encoding='utf-8'?>
<!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt">
<LIGO_LW>
  <LIGO_LW Name="psd">
    <Time Type="GPS" Name="epoch">1000000000</Time>
    <Param Type="lstring" Name="channel:param">X1:TEST-CHANNEL_1</Param>
    <Array Type="real_8" Name="PSD:array" Unit="Hz^-1">
      <Dim Start="0" Scale="1" Name="Frequency" Unit="Hz">10</Dim>
      <Dim Name="Frequency,Real">2</Dim>
      <Stream Delimiter=" " Type="Local">
        0 1
        1 2
        2 3
        3 4
        4 5
        5 6
        6 7
        7 8
        8 9
        9 10
      </Stream>
    </Array>
  </LIGO_LW>
  <LIGO_LW Name="psd">
    <Param Type="lstring" Name="channel:param">X1:TEST-CHANNEL_2</Param>
    <Array Type="real_8" Name="PSD:array" Unit="Hz^-1">
      <Dim Start="0" Scale="1" Name="Frequency" Unit="Hz">10</Dim>
      <Dim Name="Frequency,Real">2</Dim>
      <Stream Delimiter=" " Type="Local">
        0 10
        1 20
        2 30
        3 40
        4 50
        5 60
        6 70
        7 80
        8 90
        9 10
      </Stream>
    </Array>
  </LIGO_LW>
</LIGO_LW>
"""  # nopep8


class TestFrequencySeries(_TestSeries):
    TEST_CLASS = FrequencySeries

    # -- test properties ------------------------

    def test_f0(self, array):
        assert array.f0 is array.x0
        array.f0 = 4
        assert array.f0 == 4 * units.Hz

    def test_df(self, array):
        assert array.df is array.dx
        array.df = 4
        assert array.df == 4 * units.Hz

    def test_frequencies(self, array):
        assert array.frequencies is array.xindex
        utils.assert_quantity_equal(
            array.frequencies, numpy.arange(array.size) * array.df + array.f0)

    # -- test methods ---------------------------

    def test_plot(self, array):
        with rc_context(rc={'text.usetex': False}):
            plot = array.plot()
            line = plot.gca().lines[0]
            utils.assert_array_equal(line.get_xdata(), array.xindex.value)
            utils.assert_array_equal(line.get_ydata(), array.value)
            plot.save(BytesIO(), format='png')
            plot.close()

    def test_ifft(self):
        # construct a TimeSeries, then check that it is unchanged by
        # the operation TimeSeries.fft().ifft()
        ts = TimeSeries([1.0, 0.0, -1.0, 0.0], sample_rate=1.0)
        assert ts.fft().ifft() == ts
        utils.assert_allclose(ts.fft().ifft().value, ts.value)

    def test_filter(self, array):
        a2 = array.filter([100], [1], 1e-2)
        assert isinstance(a2, type(array))
        utils.assert_quantity_equal(a2.frequencies, array.frequencies)

        # manually rebuild the filter to test it works
        b, a, = signal.zpk2tf([100], [1], 1e-2)
        fresp = abs(signal.freqs(b, a, array.frequencies.value)[1])
        utils.assert_array_equal(a2.value, fresp * array.value)

    def test_zpk(self, array):
        a2 = array.zpk([100], [1], 1e-2)
        assert isinstance(a2, type(array))
        utils.assert_quantity_equal(a2.frequencies, array.frequencies)

    def test_inject(self):
        # create a timeseries out of an array of zeros
        df, nyquist = 1, 2048
        nsamp = int(nyquist/df) + 1
        data = FrequencySeries(numpy.zeros(nsamp), f0=0, df=df, unit='')

        # create a second timeseries to inject into the first
        w_nyquist = 1024
        w_nsamp = int(w_nyquist/df) + 1
        sig = FrequencySeries(numpy.ones(w_nsamp), f0=0, df=df, unit='')

        # test that we recover this waveform when we add it to data,
        # and that the operation does not change the original data
        new_data = data.inject(sig)
        assert new_data.unit == data.unit
        assert new_data.size == data.size
        ind, = new_data.value.nonzero()
        assert len(ind) == sig.size
        utils.assert_allclose(new_data.value[ind], sig.value)
        utils.assert_allclose(data.value, numpy.zeros(nsamp))

    def test_interpolate(self):
        # create a simple FrequencySeries
        df, nyquist = 1, 256
        nsamp = int(nyquist/df) + 1
        fseries = FrequencySeries(numpy.ones(nsamp), f0=0, df=df, unit='')

        # create an interpolated FrequencySeries
        newf = fseries.interpolate(df/2.)

        # check that the interpolated series is what was expected
        assert newf.unit == fseries.unit
        assert newf.size == 2*(fseries.size - 1) + 1
        utils.assert_allclose(newf.value, numpy.ones(2*int(nyquist/df) + 1))

    @utils.skip_missing_dependency('lal')
    def test_to_from_lal(self, array):
        import lal

        array.epoch = 0

        # check that to + from returns the same array
        lalts = array.to_lal()
        a2 = type(array).from_lal(lalts)
        utils.assert_quantity_sub_equal(array, a2, exclude=['name', 'channel'])
        assert a2.name == array.name

        # test copy=False
        a2 = type(array).from_lal(lalts, copy=False)
        assert shares_memory(a2.value, lalts.data.data)

        # test units
        array.override_unit('undef')
        with pytest.warns(UserWarning):
            lalts = array.to_lal()
        assert lalts.sampleUnits == lal.DimensionlessUnit
        a2 = self.TEST_CLASS.from_lal(lalts)
        assert a2.unit is units.dimensionless_unscaled

    @utils.skip_missing_dependency('lal')
    @utils.skip_missing_dependency('pycbc')
    def test_to_from_pycbc(self, array):
        from pycbc.types import FrequencySeries as PyCBCFrequencySeries

        array.epoch = 0

        # test default conversion
        pycbcfs = array.to_pycbc()
        assert isinstance(pycbcfs, PyCBCFrequencySeries)
        utils.assert_array_equal(array.value, pycbcfs.data)
        assert array.f0.value == 0 * units.Hz
        assert array.df.value == pycbcfs.delta_f
        assert array.epoch.gps == pycbcfs.epoch

        # go back and check we get back what we put in in the first place
        a2 = type(array).from_pycbc(pycbcfs)
        utils.assert_quantity_sub_equal(
            array, a2, exclude=['name', 'unit', 'channel'])

        # test copy=False
        a2 = type(array).from_pycbc(array.to_pycbc(copy=False), copy=False)
        assert shares_memory(array.value, a2.value)

    @pytest.mark.parametrize('format', [
        'txt',
        'csv',
    ])
    def test_read_write(self, array, format):
        utils.test_read_write(
            array, format,
            assert_equal=utils.assert_quantity_sub_equal,
            assert_kw={'exclude': ['name', 'channel', 'unit', 'epoch']})

    @utils.skip_missing_dependency('glue.ligolw.utils.ligolw_add')
    def test_read_ligolw(self):
        with tempfile.NamedTemporaryFile(mode='w+') as fobj:
            fobj.write(LIGO_LW_ARRAY)
            array = FrequencySeries.read(
                fobj, 'psd', match={'channel': 'X1:TEST-CHANNEL_1'})
            utils.assert_array_equal(array, list(range(1, 11)) / units.Hz)
            utils.assert_array_equal(array.frequencies,
                                     list(range(10)) * units.Hz)
            assert numpy.isclose(array.epoch.gps, 1000000000)  # precision gah!
            assert array.unit == units.Hz ** -1

            array2 = FrequencySeries.read(
                fobj, 'psd', match={'channel': 'X1:TEST-CHANNEL_2'})
            assert array2.epoch is None

            # assert errors
            with pytest.raises(ValueError):
                FrequencySeries.read(fobj, 'blah')
            with pytest.raises(ValueError):
                FrequencySeries.read(fobj, 'psd')
            with pytest.raises(ValueError):
                FrequencySeries.read(
                    fobj, 'psd',
                    match={'channel': 'X1:TEST-CHANNEL_1', 'blah': 'blah'})
